"""
Linear profiles
===============
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains classes to compute number density and charge
density profiles and related quantities, like surface charge densities
and potential profiles.
"""

import logging
from numbers import Real
from typing import Any, Union
import warnings

import MDAnalysis as mda
import numpy as np
from scipy import integrate, sparse

from .base import Hash, DynamicAnalysisBase
from .. import FOUND_OPENMM, Q_, ureg
from ..algorithm.accelerated import numba_histogram_bin_edges, numba_histogram
from ..algorithm.molecule import center_of_mass
from ..algorithm.topology import unwrap, wrap
from ..algorithm.unit import is_unitless, strip_unit

if FOUND_OPENMM:
    from openmm import unit

_ELECTROSTATIC_CONVERSION_FACTORS = (
    (1 * ureg.elementary_charge / (ureg.vacuum_permittivity * ureg.angstrom)).m_as(
        ureg.volt
    ),
    4 * np.pi,
)


def calculate_surface_charge_density(
    bins: np.ndarray[float],
    charge_density_profile: np.ndarray[float],
    dielectric: float = None,
    *,
    L: float = None,
    dV: Union[float, np.ndarray[float]] = None,
    reduced: bool = False,
) -> Union[float, np.ndarray[float]]:
    r"""
    Calculates the surface charge density :math:`\sigma_q` using the
    charge density profile :math:`\rho_q(z)`.

    The surface charge density is given by

    .. math::

       \sigma_q=\frac{1}{L}\left(
       \varepsilon_0\varepsilon_\mathrm{r}\Delta V
       -\int_0^L z\rho_q(z)\,\mathrm{d}z\right)

    where :math:`\varepsilon_0` and :math:`\varepsilon_\mathrm{r}` are
    the vacuum and relative permittivities, respectively,
    :math:`\Delta V` is the potential difference, and :math:`L` is the
    system dimension.

    The first (static) term accounts for the applied potential across
    the dielectric medium, while the second (polarization) term accounts
    for the distribution of charged species.

    Parameters
    ----------
    bins : array-like
        Histogram bin centers :math:`z` for the charge density profile
        in `charge_density_profile`.

        **Shape**: :math:`(N_\mathrm{bins},)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    charge_density_profile : array-like
        Charge density profile :math:`\rho_q(z)`.

        **Shape**: :math:`(N_\mathrm{bins},)` or
        :math:`(N_\mathrm{frames},\,N_\mathrm{bins})`.

        **Reference unit**: :math:`\mathrm{e/Å}^{-3}`.

    dielectric : `float`, optional
        Relative permittivity or static dielectric constant
        :math:`\varepsilon_\mathrm{r}`. Required if `dV` is provided.

    L : `float`, keyword-only, optional
        System size :math:`L` in the dimension that `bins` and
        `charge_density_profile` were calculated in. If not specified,
        it is determined using the first and last values in `bins`,
        assuming that the bin centers are uniformly spaced.

        **Reference unit**: :math:`\mathrm{Å}`.

    dV : `float` or array-like, keyword-only, optional
        Potential difference :math:`\Delta V` across the system
        dimension specified in `axis`. If not provided, only the
        polarizable part (second term) of the surface charge density is
        calculated.

        **Shape**: Scalar or :math:`(N_\mathrm{frames},)`.

        **Reference unit**: :math:`\mathrm{V}`.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units.

    Returns
    -------
    surface_charge_density : `numpy.ndarray`
        Surface charge density :math:`\sigma_q`.

        **Shape**: Scalar or :math:`(N_\mathrm{frames},)`.

        **Reference unit**: :math:`\mathrm{e/Å}^2`.
    """

    # Check input array shapes
    if bins.ndim != 1:
        raise ValueError("'bins' must be a one-dimensional array.")
    if charge_density_profile.ndim not in {1, 2}:
        emsg = "'charge_density_profile' must be a one- or " "two-dimensional array."
        raise ValueError(emsg)
    if bins.shape[0] != charge_density_profile.shape[-1]:
        emsg = (
            "'bins' and 'charge_density_profile' must have the same"
            "length in the last dimension."
        )
        raise ValueError(emsg)
    if dV is not None:
        if dielectric is None:
            raise ValueError("'dielectric' must be specified when 'dV' is.")
        if not isinstance(dV, Real) and dV.shape[0] != charge_density_profile.shape[0]:
            emsg = (
                "The number of potential differences in 'dVs' must "
                "be equal to the number of frames in "
                "'charge_density_profile'."
            )
            raise ValueError(emsg)

    # Determine system size, if not provided
    dz = bins[1] - bins[0]
    if (uniform := np.allclose(np.diff(bins), dz)) and not np.isclose(bins[0], dz / 2):
        wmsg = (
            "'bins' currently does not start at zero, so it will "
            f"be shifted left by {(shift := bins[0] - dz / 2):.6g}."
        )
        warnings.warn(wmsg)
        bins -= shift
    if L is None:
        if not uniform:
            emsg = "'L' must be specified when 'bins' is not uniformly spaced."
            raise ValueError(emsg)
        L = bins[-1] - bins[0] + dz

    y = -integrate.trapezoid(bins * charge_density_profile, bins)
    if dV is not None:
        y += dielectric * dV / _ELECTROSTATIC_CONVERSION_FACTORS[reduced]
    return y / L


def calculate_potential_profile(
    bins: np.ndarray[float],
    charge_density_profile: np.ndarray[float],
    dielectric: float,
    *,
    L: float = None,
    sigma_q: Union[float, np.ndarray[float]] = None,
    dV: Union[float, np.ndarray[float]] = None,
    threshold: float = 1e-5,
    V0: Union[float, np.ndarray[float]] = 0,
    method: str = "integral",
    pbc: bool = False,
    reduced: bool = False,
) -> np.ndarray[float]:
    r"""
    Calculates the potential profile :math:`\Psi(z)` using the charge
    density profile :math:`\rho_q(z)` by numerically solving Poisson's
    equation for electrostatics.

    Poisson's equation is given by

    .. math::

       \varepsilon_0\varepsilon_\mathrm{r}\nabla^2\Psi(z)=-\rho_q(z)

    where :math:`\varepsilon_0` and :math:`\varepsilon_\mathrm{r}` are
    the vacuum and relative permittivities, respectively.

    The boundary conditions (BCs) are

    .. math::

       \left.\frac{\partial\Psi}{\partial z}\right\rvert_{z=0}&
       =-\frac{\sigma_q}{\varepsilon_0\varepsilon_\mathrm{r}}\\
       \left.\Psi\right\rvert_{z=0}&=\Psi_0

    The first BC is used to ensure that the electric field in the bulk
    of the solution is zero, while the second BC is used to set the
    potential on the left electrode.

    Poisson's equation can be evaluated by using the trapezoidal rule
    to numerically integrate the charge density profile twice:

    1. Integrate the charge density profile.
    2. Apply the first BC by subtracting :math:`\sigma_q` from all points.
    3. Integrate the profile from Step 2.
    4. Divide by :math:`-\varepsilon_0\varepsilon_\mathrm{r}`.
    5. Apply the second BC by adding :math:`\Psi_0` to all points.

    This method is fast but requires many histogram bins to accurately
    resolve the potential profile.

    Alternatively, Poisson's equation can be written as a system of
    linear equations

    .. math::

       A\mathbf{\Psi}=\mathbf{b}

    using second-order finite differences. :math:`A` is a tridiagonal
    matrix and :math:`\mathbf{b}` is a vector containing the charge
    density profile, with boundary conditions applied in the first and
    last rows.

    The inner equations are given by

    .. math::

       \Psi_i^{''}=\frac{\Psi_{i-1}-2\Psi_i+\Psi_{i+1}}{h^2}
       =-\frac{1}{\varepsilon_0\varepsilon_\mathrm{r}}\rho_{q,\,i}

    where :math:`i` is the bin index and :math:`h` is the bin width.

    In the case of periodic boundary conditions, the first and last
    equations are given by

    .. math::

       \Psi_0^{''}&=\frac{\Psi_{N-1}-2\Psi_0+\Psi_1}{h^2}
       =-\frac{1}{\varepsilon_0\varepsilon_\mathrm{r}}\rho_{q,\,0}\\
       \Psi_{N-1}^{''}&=\frac{\Psi_{N-2}-2\Psi_{N-1}+\Psi_0}{h^2}
       =-\frac{1}{\varepsilon_0\varepsilon_\mathrm{r}}\rho_{q,\,N-1}

    When the system has slab geometry, the boundary conditions are
    implemented via

    .. math::

       \Psi_0&=\Psi_0\\
       \Psi_0^\prime&=\frac{-3\Psi_0+4\Psi_1-\Psi_2}{2h}
       =\frac{\sigma_q}{\varepsilon_0\varepsilon_\mathrm{r}}

    This method is slower but can be more accurate even with fewer
    histogram bins for bulk systems with periodic boundary conditions.

    Parameters
    ----------
    bins : array-like
        Histogram bin centers :math:`z` for the charge density profile
        in `charge_density_profile`.

        **Shape**: :math:`(N_\mathrm{bins},)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    charge_density_profile : array-like
        Charge density profile :math:`\rho_q(z)`.

        **Shape**: :math:`(N_\mathrm{bins},)` or
        :math:`(N_\mathrm{frames},\,N_\mathrm{bins})`.

        **Reference unit**: :math:`\mathrm{e/Å}^{-3}`.

    dielectric : `float`
        Relative permittivity or static dielectric constant
        :math:`\varepsilon_\mathrm{r}`.

    L : `float`, keyword-only, optional
        System size in the dimension that `bins` and
        `charge_density_profile` were calculated in. If not specified,
        it is determined using the first and last values in `bins`,
        assuming that the bin centers are uniformly spaced.

        **Reference unit**: :math:`\mathrm{Å}`.

    sigma_q : `float` or array-like, keyword-only, optional
        Surface charge density :math:`\sigma_q`. Used as a boundary
        condition. If not provided, it is determined using `dV` and
        `charge_density_profile`. If `dV` is also not provided, the
        average value in the center of the integrated charge density
        profile is used if :code:`method="integral"`.

        .. note::

           :math:`\sigma_q` and :math:`\Delta\Psi` should have the same
           sign.

        **Shape**: Scalar or :math:`(N_\mathrm{frames},)`.

        **Reference unit**: :math:`\mathrm{e/Å^2}`.

    dV : `float` or array-like, keyword-only, optional
        Potential difference :math:`\Delta\Psi` across the system
        dimension specified in `axis`. Only used if `sigma_q` was not
        provided.

        **Shape**: Scalar or :math:`(N_\mathrm{frames},)`.

        **Reference unit**: :math:`\mathrm{V}`.

    V0 : `float` or array-like, keyword-only, default: :code:`0`
        Potential :math:`\Psi_0` at the left boundary.

        **Shape**: Scalar or :math:`(N_\mathrm{frames},)`.

        **Reference unit**: :math:`\mathrm{V}`.

    threshold : `float`, keyword-only, default: :code:`1e-5`
        Threshold for determining the plateau regions in the centers of
        the integrated charge density profiles to be used as estimates
        of `sigma_q`. Only used if `sigma_q` was not provided and
        cannot be calculated using `dV` and `charge_density_profile`.

    method : `str`, keyword-only, default: :code:`"integral"`
        Method used to calculate the potential profiles.

        **Valid values**: :code:`"integral"`, :code:`"matrix"`.

    pbc : `bool`, keyword-only, default: :code:`False`
        Specifies whether the axis has periodic boundary conditions.
        Only used when :code:`method="matrix"`.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units.

    Returns
    -------
    potential : `numpy.ndarray`
        Potential profile :math:`\Psi(z)`.

        **Shape**: Same as `charge_density_profile`.

        **Reference unit**: :math:`\mathrm{V}`.
    """

    ecf = _ELECTROSTATIC_CONVERSION_FACTORS[reduced]

    # Check input array shapes
    if bins.ndim != 1:
        raise ValueError("'bins' must be a one-dimensional array.")
    if charge_density_profile.ndim not in {1, 2}:
        emsg = "'charge_density_profile' must be a one- or " "two-dimensional array."
        raise ValueError(emsg)
    if bins.shape[0] != charge_density_profile.shape[-1]:
        emsg = (
            "'bins' and 'charge_density_profile' must have the same"
            "length in the last dimension."
        )
        raise ValueError(emsg)
    if (
        sigma_q is not None
        and not isinstance(sigma_q, Real)
        and sigma_q.shape[0] != charge_density_profile.shape[0]
    ):
        emsg = (
            "The number of surface charge densities in 'sigmas_q' "
            "must be equal to the number of frames in "
            "'charge_density_profile'."
        )
        raise ValueError(emsg)
    if (
        dV is not None
        and not isinstance(dV, Real)
        and dV.shape[0] != charge_density_profile.shape[0]
    ):
        emsg = (
            "The number of potential differences in 'dVs' must be "
            "equal to the number of frames in "
            "'charge_density_profile'."
        )
        raise ValueError(emsg)
    if (
        V0 is not None
        and not isinstance(V0, Real)
        and V0.shape[0] != charge_density_profile.shape[0]
    ):
        emsg = (
            "The number of potentials in 'V0s' must be equal to "
            "the number of frames in 'charge_density_profile'."
        )
        raise ValueError(emsg)

    # Determine system size, if not provided
    dz = bins[1] - bins[0]
    if (uniform := np.allclose(np.diff(bins), dz)) and not np.isclose(bins[0], dz / 2):
        wmsg = (
            "'bins' currently does not start at zero, so it will "
            f"be shifted left by {(shift := bins[0] - dz / 2):.6g}."
        )
        warnings.warn(wmsg)
        bins -= shift
    if L is None:
        if not uniform:
            emsg = "'L' must be specified when 'bins' is not uniformly spaced."
            raise ValueError(emsg)
        L = bins[-1] - bins[0] + dz

    # Calculate surface charge density, if not provided
    if sigma_q is None and dV is not None:
        sigma_q = (
            dielectric * dV / ecf
            - integrate.trapezoid(bins * charge_density_profile, bins)
        ) / L

    if method == "integral":
        potential = integrate.cumulative_trapezoid(
            charge_density_profile, bins, initial=0
        )

        if sigma_q is None:
            wmsg = (
                "No surface charge density information. The value "
                "will be extracted from the integrated charge "
                "density profile, which may be inaccurate due to "
                "numerical errors."
            )
            warnings.warn(wmsg)

            # Get surface charge density from the integrated charge
            # density profile
            cut_indices = (
                np.where(
                    np.diff(
                        np.abs(
                            np.gradient(
                                potential
                                if potential.ndim == 1
                                else potential.mean(axis=0)
                            )
                        )
                        < threshold
                    )
                )[0]
                + 1
            )
            if len(cut_indices) == 0:
                logging.warning(
                    "No bulk plateau region found in the integrated "
                    "charge density profile. The average value over "
                    "the entire profile will be used."
                )
                sigma_q = potential.mean(axis=-1, keepdims=True)
            else:
                target_index = potential.shape[-1] // 2
                if potential.ndim == 1:
                    sigma_q = potential[
                        cut_indices[cut_indices <= target_index][-1] : cut_indices[
                            cut_indices >= target_index
                        ][0]
                    ].mean()
                else:
                    sigma_q = potential[
                        :,
                        cut_indices[cut_indices <= target_index][-1] : cut_indices[
                            cut_indices >= target_index
                        ][0],
                    ].mean(axis=-1, keepdims=True)

        return (
            V0
            - ecf
            * (integrate.cumulative_trapezoid(potential - sigma_q, bins, initial=0))
            / dielectric
        )

    elif method == "matrix":
        if sigma_q is None:
            emsg = (
                "No surface charge density information. Either "
                "'sigma_q' or 'dV' must be provided when "
                "method='matrix'."
            )
            raise ValueError(emsg)

        if not np.allclose(np.diff(bins), dz):
            raise ValueError("'bins' must be uniformly spaced.")

        # Set up matrix and load vector for second-order finite
        # difference method
        N = len(bins)
        A = sparse.diags((1, -2, 1), (-1, 0, 1), shape=(N, N), format="csc")
        b = charge_density_profile.copy().T
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=sparse.SparseEfficiencyWarning)

            # Apply boundary conditions and solve
            if pbc:
                A[0, -1] = A[-1, 0] = 1
                b *= -ecf * dz**2 / dielectric
                psi = np.empty_like(b)
                psi[1:] = sparse.linalg.spsolve(A[1:, 1:], b[1:])
                psi[0] = psi[-1]
                return psi
            else:
                A[0, :3] = -1.5, 2, -0.5
                A[-1, 0] = 1
                A[-1, -2:] = 0
                b[0] = ecf * dz * sigma_q / dielectric
                b[1:-1] *= -ecf * dz**2 / dielectric
                b[-1] = V0
                return sparse.linalg.spsolve(A, b)


class DensityProfile(DynamicAnalysisBase):
    """
    Serial and parallel implementations to calculate the number and
    charge density profiles :math:`\\rho_i(z)` and :math:`\\rho_q(z)` of
    a constant-volume system along the specified axes.

    The microscopic number density profile of species :math:`i` is
    calculated by binning particle positions along an axis :math:`z`
    using

    .. math::

       \\rho_i(z)=\\frac{V}{N_\\mathrm{bins}}\\left\\langle
       \\sum_\\alpha\\delta(z-z_\\alpha)\\right\\rangle

    where :math:`V` is the system volume and :math:`N_\\mathrm{bins}` is
    the number of bins. The angular brackets denote an ensemble average.

    If the species carry charges, the charge density profile can be
    obtained using

    .. math::

       \\rho_q(z)=e\\sum_i z_i\\rho_i(z)

    where :math:`z_i` is the charge number of species :math:`i` and
    :math:`e` is the elementary charge.

    With the charge density profile, the surface charge density is
    given by

    .. math::

       \\sigma_q=\\frac{1}{L}\\left(
       \\varepsilon_0\\varepsilon_\\mathrm{r}\\Delta\\Psi
       -\\int_0^L z\\rho_q(z)\\,\\mathrm{d}z\\right)

    where :math:`\\varepsilon_0` and :math:`\\varepsilon_\\mathrm{r}`
    are the vacuum and relative permittivities, respectively,
    :math:`\\Delta\\Psi` is the potential difference, and :math:`L` is
    the system dimension, and the potential profile can be computed by
    numerically solving Poisson's equation for electrostatics:

    .. math::

       \\varepsilon_0\\varepsilon_\\mathrm{r}\\nabla^2\\Psi(z)=-\\rho_q(z)

    Parameters
    ----------
    groups : `MDAnalysis.AtomGroup` or array-like
        Groups of atoms for which density profiles are calculated.

    groupings : `str` or array-like, default: :code:`"atoms"`
        Determines whether the centers of mass are used in lieu of
        individual atom positions. If `groupings` is a `str`, the same
        value is used for all `groups`.

        .. note::

           If the desired grouping is not :code:`"atoms"`,

           * the trajectory file should have segments (or chains)
             containing residues (or molecules) and residues containing
             atoms, and

           * residues and segments should be locally unwrapped at the
             simulation box edges, if not already, using
             :class:`MDAnalysis.transformations.wrap.unwrap`,
             :meth:`MDAnalysis.core.groups.AtomGroup.unwrap`, or
             :func:`MDAnalysis.lib.mdamath.make_whole`.

        .. container::

           **Valid values**:

           * :code:`"atoms"`: Atom positions (generally or for
             coarse-grained simulations).
           * :code:`"residues"`: Residues' centers of mass (for
             atomistic simulations).
           * :code:`"segments"`: Segments' centers of mass (for
             atomistic polymer simulations).

    axes : `int`, `str`, or array-like, default: :code:`"xyz"`
        Axes along which to compute the density profiles.

        .. container::

           **Examples**:

           * :code:`2` for the :math:`z`-direction.
           * :code:`"xy"` for the :math:`x`- and :math:`y`-directions.
           * :code:`(0, 1)` for the :math:`x`- and :math:`y`-directions.

    n_bins : `int` or array-like
        Number of histogram bins :math:`N_\\mathrm{bins}` for each axis
        in `axes`. If an `int` is provided, the same value is used for
        all axes.

    charges : array-like, `openmm.unit.Quantity`, or `pint.Quantity`, \
    keyword-only, optional
        Charges :math:`q_i` for the entities in the 
        :math:`N_\\mathrm{groups}` atom groups in `groups`. If not
        provided, they will be retrieved from the main
        :class:`MDAnalysis.core.universe.Universe` object only if it
        contains charge information.

        .. note::

           Depending on the grouping for a specific atom group, all
           entities (atoms, residues, or segments) must carry the same
           charge. Otherwise, the charge density contribution for that
           atom group would not make sense. If this condition does not
           hold, change how the atoms are grouped in the atom groups so
           that all entities share the same charge.

        **Shape**: :math:`(N_\\mathrm{groups},)`.

        **Reference unit**: :math:`\\mathrm{e}`.

    dimensions : array-like, `openmm.unit.Quantity`, or \
    `pint.Quantity`, keyword-only, optional
        System dimensions :math:`(L_x,\\,L_y,\\,L_z)`. If the
        :class:`MDAnalysis.core.universe.Universe` object that the
        atom groups in `groups` belong to does not contain
        dimensionality information, provide it here. Affected by
        `dim_scales`.

        **Shape**: :math:`(3,)`.

        **Reference unit**: :math:`\\mathrm{Å}`.

    dim_scales : `float` or array-like, keyword-only, optional
        Scale factors for the system dimensions. If an `int` is
        provided, the same value is used for all axes.

        **Shape**: :math:`(3,)`.

    average : `bool`, keyword-only, default: :code:`True`
        Determines whether the density profiles are averaged over the
        :math:`N_\\mathrm{frames}` analysis frames.

    recenter : `int`, `list`, `MDAnalysis.AtomGroup`, or `tuple`, \
    keyword-only, optional
        Constrains the center of mass of an atom group by adjusting the
        particle coordinates every analysis frame. Either specify an
        :class:`MDAnalysis.core.groups.AtomGroup`, its index within
        `groups`, a list of atom groups or their indices, or a tuple
        containing the aforementioned information and the fixed center
        of mass coordinates, in that order. To avoid recentering in a
        specific dimension, set the coordinate to :code:`numpy.nan`. If
        the center of mass is not specified, the center of the
        simulation box is used.

        **Shape**: :math:`(3,)` for the fixed center of mass.

    reduced : `bool`, keyword-only, default: :code:`False`
        Specifies whether the data is in reduced units. Affects
        `results.number_densities`, `results.charge_densities`, etc.

    parallel : `bool`, keyword-only, default: :code:`False`
        Determines whether the analysis is performed in parallel.

        .. note::

           The Joblib threading backend generally provides the best
           performance.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    **kwargs
        Additional keyword arguments to pass to
        :class:`MDAnalysis.analysis.base.AnalysisBase`.

    Attributes
    ----------
    universe : `MDAnalysis.Universe`
        :class:`MDAnalysis.core.universe.Universe` object containing all
        information describing the simulation system.

    axes : `tuple`
        Axes along which the density profiles are calculated.

    results.units : `dict`
        Reference units for the results. For example, to get the
        reference units for `results.bins`, call
        :code:`results.units["bins"]`.

    results.times : `numpy.ndarray`
        Times :math:`t`. Only available if :code:`average=False`.

        **Shape**: :math:`(N_\\mathrm{frames},)`.

        **Reference unit**: :math:`\\mathrm{ps}`.

    results.bins : `dict`
        Bin centers :math:`z` corresponding to the density profiles in
        each dimension. The key is the axis, e.g.,
        :code:`results.bins["z"]` for the :math:`z`-axis.

        **Shape**: Each array has shape :math:`(N_\\mathrm{bins},)`.

        **Reference unit**: :math:`\\mathrm{Å}`.

    results.bin_edges : `dict`
        Bin edges corresponding to the density profiles in each
        dimension. The key is the axis, e.g.,
        :code:`results.bin_edges["z"]` for the :math:`z`-axis.

        **Shape**: Each array has shape :math:`(N_\\mathrm{bins}+1,)`.

        **Reference unit**: :math:`\\mathrm{Å}`.

    results.number_densities : `dict`
        Number density profiles :math:`\\rho(z)`. The key is the axis,
        e.g., :code:`results.number_densities["z"]` for the
        :math:`z`-axis.

        **Shape**: Each array has shape
        :math:`(N_\\mathrm{groups},\\,N_\\mathrm{bins})`. If
        :code:`average=False`, an additional second dimension of
        length :math:`N_\\mathrm{frames}` is present.

        **Reference unit**: :math:`\\mathrm{Å}^{-3}`.

    results.charge_densities : `dict`
        Charge density profiles :math:`\\rho_q(z)`. Only available if
        charge information was found or provided. The key is the axis,
        e.g., :code:`results.charge_densities["z"]` for the
        :math:`z`-axis.

        **Shape**: Each array has shape :math:`(N_\\mathrm{bins},)`. If
        :code:`average=False`, an additional first dimension of length
        :math:`N_\\mathrm{frames}` is present.

        **Reference unit**: :math:`\\mathrm{e/Å}^{-3}`.

    results.surface_charge_densities : `numpy.ndarray`
        Surface charge densities :math:`\\sigma_q`. Only available after
        running :meth:`calculate_surface_charge_densities`.

        **Shape**: :math:`(N_\\mathrm{axes},)` or
        :math:`(N_\\mathrm{axes},\\,N_\\mathrm{frames})`.

    results.potentials : `dict`
        Potential profiles :math:`\\Psi(z)`. Only available after
        running :meth:`calculate_potential_profiles`. The key is the
        axis, e.g., :code:`results.potentials["z"]` for the
        :math:`z`-axis.

        **Shape**: Each array has shape :math:`(N_\\mathrm{bins},)`. If
        :code:`average=False`, an additional second dimension of
        length :math:`N_\\mathrm{frames}` is present.

        **Reference unit**: :math:`\\mathrm{V}`.
    
    Example
    --------
    First, this analysis class must be imported:

    >>> from mdcraft.analysis.profile import DensityProfile

    Then, after loading a simulation trajectory:

    >>> universe = mda.Universe("simulation.nc", "topology.cif")

    We must then select the atom-groups to be analyzed:

    >>> ag1 = universe.select_atoms("resname CAT")
    >>> ag2 = universe.select_atoms("resname ANI")
    
    The `DensityProfile` class can be instantiated with the selected atom-groups and the axes to calculate the density profiles along (charges can also be provided to obtain the charge density profiles):
    
    >>> prof = DensityProfile([ag1, ag2], groupings="residues", axes="x", charges=[1, -1])
    >>> prof.run()

    The results can be obtained under the `results` attribute:
    
    >>> prof.results.number_densities["x"]
    
    To calculate the surface charge densities, the `calculate_surface_charge_densities` method can be used:
    
    >>> prof.calculate_surface_charge_densities(dV=1, dielectric=80)
    
    Further, to calculate the potential profiles, the `calculate_potential_profiles` method can be used:
    
    >>> prof.calculate_potential_profiles(dielectric=80)
    
    These results can be saved to a file using the `save` method:
        
    >>> prof.save("density_profiles")
    """

    def __init__(
        self,
        groups: Union[mda.AtomGroup, tuple[mda.AtomGroup]],
        groupings: Union[str, tuple[str]] = "atoms",
        axes: Union[int, str, tuple[Union[int, str]]] = "xyz",
        n_bins: Union[int, tuple[int]] = 201,
        *,
        charges: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
        dimensions: Union[np.ndarray[float], "unit.Quantity", Q_] = None,
        recenter: Union[
            mda.AtomGroup,
            int,
            list[mda.AtomGroup, int],
            tuple[
                Union[mda.AtomGroup, int, list[mda.AtomGroup, int]], np.ndarray[float]
            ],
        ] = None,
        dim_scales: Union[float, tuple[float]] = 1,
        average: bool = True,
        reduced: bool = False,
        parallel: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> None:

        self._groups = [groups] if isinstance(groups, mda.AtomGroup) else groups
        self.universe = self._groups[0].universe
        super().__init__(self.universe.trajectory, parallel, verbose, **kwargs)

        GROUPINGS = {"atoms", "residues", "segments"}
        self._n_groups = len(self._groups)
        if isinstance(groupings, str):
            if groupings not in GROUPINGS:
                emsg = (
                    f"Invalid grouping '{groupings}'. Valid values: "
                    "'" + "', '".join(GROUPINGS) + "'."
                )
                raise ValueError(emsg)
            self._groupings = self._n_groups * [groupings]
        else:
            if self._n_groups != len(groupings):
                emsg = (
                    "The shape of 'groupings' is incompatible with " "that of 'groups'."
                )
                raise ValueError(emsg)
            for gr in groupings:
                if gr not in GROUPINGS:
                    emsg = (
                        f"Invalid grouping '{gr}'. Valid values: "
                        "'" + "', '".join(GROUPINGS) + "'."
                    )
                    raise ValueError(emsg)
            self._groupings = groupings

        if isinstance(axes, int):
            self._axis_indices = np.array((axes,), dtype=int)
        else:
            self._axis_indices = np.fromiter(
                (ord(ax.lower()) - 120 if isinstance(ax, str) else ax for ax in axes),
                count=len(axes),
                dtype=int,
            )
        self.axes = tuple(chr(120 + i) for i in self._axis_indices)
        if not all(ax in "xyz" for ax in self.axes):
            raise ValueError("Invalid axis passed in 'axes'.")
        self._n_axes = len(self.axes)

        if isinstance(n_bins, int):
            self._n_bins = n_bins * np.ones(len(self.axes), dtype=int)
        elif isinstance(n_bins, str):
            emsg = "'n_bins' must be an integer or an iterable object."
            raise ValueError(emsg)
        else:
            if len(n_bins) != len(self.axes):
                emsg = (
                    "The shape of 'n_bins' is incompatible with "
                    "the number of axes to calculate density "
                    "profiles along."
                )
                raise ValueError(emsg)
            if not all(isinstance(n, int) for n in n_bins):
                emsg = "All bin counts in 'n_bins' must be integers."
                raise ValueError(emsg)
            self._n_bins = n_bins

        if charges is not None:
            if len(charges) != self._n_groups:
                emsg = (
                    "The shape of 'charges' is incompatible with " "that of 'groups'."
                )
                raise ValueError(emsg)
            if reduced and not is_unitless(charges):
                emsg = "'charges' cannot have units when 'reduced=True'."
                raise TypeError(emsg)
            self._charges = np.asarray(strip_unit(charges, "e")[0])
        elif hasattr(self.universe.atoms, "charges"):
            self._charges = np.empty(self._n_groups)
            for i, (ag, gr) in enumerate(zip(self._groups, self._groupings)):
                qs = getattr(ag, gr).charges
                if not np.allclose(qs, q := qs[0]):
                    self._charges = None
                    wmsg = (
                        f"Not all {gr} in `groups[{i}]` share the "
                        "same charge. The charge density profile will "
                        "not be calculated."
                    )
                    warnings.warn(wmsg)
                    break
                self._charges[i] = q
        else:
            self._charges = None

        if dimensions is not None:
            if len(dimensions) != 3:
                raise ValueError("'dimensions' must have length 3.")
            if reduced and not is_unitless(dimensions):
                emsg = "'dimensions' cannot have units when 'reduced=True'."
                raise TypeError(emsg)
            self._dimensions = np.asarray(strip_unit(dimensions, "Å")[0])
        elif self.universe.dimensions is not None:
            self._dimensions = self.universe.dimensions[:3].copy()
        else:
            raise ValueError("No system dimensions found or provided.")

        if isinstance(dim_scales, Real) or (
            len(dim_scales) == 3 and all(isinstance(f, Real) for f in dim_scales)
        ):
            self._dimensions *= dim_scales
        else:
            emsg = (
                "'dim_scales' must be a floating-point number "
                "or an array with shape (3,)."
            )
            raise ValueError(emsg)

        self._Ns = np.fromiter(
            (getattr(ag, f"n_{gr}") for (ag, gr) in zip(self._groups, self._groupings)),
            dtype=int,
            count=self._n_groups,
        )
        self._N = self._Ns.sum()
        self._slices = []
        _ = 0
        for N in self._Ns:
            self._slices.append(slice(_, _ + N))
            _ += N

        if recenter is None:
            self._recenter = recenter
        else:
            if isinstance(recenter, (int, mda.AtomGroup)):
                grp = recenter
                com = self._dimensions / 2
            elif isinstance(recenter, tuple) and len(recenter) == 2:
                grp, com = recenter
                if len(com) != 3:
                    emsg = (
                        "The target center of mass in 'recenter' " "must have length 3."
                    )
                    raise ValueError(emsg)
                com = np.asarray(com)
            else:
                emsg = (
                    "Invalid value passed to 'recenter'. The "
                    "argument must either be an atom group, its "
                    "index in 'groups', multiple groups/indices, or "
                    "a tuple containing the aforementioned "
                    "information and a target center of mass, in "
                    "that order."
                )
                raise ValueError(emsg)

            if isinstance(grp, int):
                if not 0 <= grp < self._n_groups:
                    emsg = "Invalid group index passed to 'recenter'."
                    raise ValueError(emsg)
                grp = self._slices[grp]
            elif isinstance(grp, mda.AtomGroup):
                try:
                    grp = self._slices[self._groups.index(grp)]
                except ValueError:
                    emsg = "The atom group passed to 'recenter' is not " "in 'groups'."
                    raise ValueError(emsg)
            else:
                try:
                    grp = np.r_[
                        tuple(
                            self._slices[
                                g if isinstance(g, int) else self._groups.index(g)
                            ]
                            for g in grp
                        )
                    ]
                except ValueError:
                    emsg = "Invalid atom group or index passed to 'recenter'."
                    raise ValueError(emsg)
            self._recenter = (grp, com)

        self._dielectrics = np.empty(self._n_axes)
        self._dVs = np.empty(self._n_axes)
        self._dielectrics[:] = self._dVs[:] = np.nan

        self._average = average
        self._reduced = reduced
        self._verbose = verbose

    def _prepare(self) -> None:

        # Specify bin centers and edges for each axis
        self.results.bins = Hash()
        self.results.bin_edges = Hash()
        for ia, ax, n in zip(self._axis_indices, self.axes, self._n_bins):
            self.results.bin_edges[ax] = numba_histogram_bin_edges(
                np.asarray((0.0, self._dimensions[ia])), n
            )
            self.results.bins[ax] = (
                self.results.bin_edges[ax][:-1] + self.results.bin_edges[ax][1:]
            ) / 2

        # Store entity masses for center of mass calculations
        self._masses = np.empty(self._N)
        for ag, gr, s in zip(self._groups, self._groupings, self._slices):
            self._masses[s] = getattr(ag, gr).masses

        if self._recenter is not None:

            # Navigate to first frame in analysis
            self._sliced_trajectory[0]

            # Preallocate arrays to determine the number of periodic
            # boundary crossings for each entity
            self._positions_old = np.empty((self._N, 3))
            for ag, gr, s in zip(self._groups, self._groupings, self._slices):
                self._positions_old[s] = (
                    ag.positions if gr == "atoms" else center_of_mass(ag, gr)
                )
            self._images = np.zeros((self._N, 3), dtype=int)
            self._thresholds = self._dimensions / 2

            if self._parallel:

                # Preallocate array to hold processed entity positions
                # for all frames to be analyzed in parallel
                self._positions = np.empty((self.n_frames, self._N, 3))
                for i, _ in enumerate(self._sliced_trajectory):

                    # Get raw entity positions for current frame
                    for ag, gr, s in zip(self._groups, self._groupings, self._slices):
                        self._positions[i, s] = (
                            ag.positions if gr == "atoms" else center_of_mass(ag, gr)
                        )

                    # Globally unwrap entity positions for correct
                    # center of mass
                    unwrap(
                        self._positions[i],
                        self._positions_old,
                        self._dimensions,
                        thresholds=self._thresholds,
                        images=self._images,
                    )

                    # Shift entity positions by the difference between
                    # the group and target centers of mass
                    self._positions[i] -= np.fromiter(
                        (
                            0 if np.isnan(tx) else gx - tx
                            for gx, tx in zip(
                                center_of_mass(
                                    positions=self._positions[i, self._recenter[0]],
                                    masses=self._masses[self._recenter[0]],
                                ),
                                self._recenter[1],
                            )
                        ),
                        dtype=float,
                        count=3,
                    )

                # Wrap entity positions back into the simulation box so that
                # they belong to a histogram bin
                wrap(self._positions, self._dimensions)

        # Preallocate arrays to hold entity positions for a given frame
        # (so that one doesn't have to be recreated each frame) and
        # number density profiles for serial analysis
        if not self._parallel:
            self._positions = np.empty((self._N, 3))
            shape = [self._n_groups]
            if not self._average:
                shape.append(self.n_frames)
            self.results.number_densities = Hash(
                {ax: np.zeros((*shape, n)) for ax, n in zip(self.axes, self._n_bins)}
            )

        # Store reference units
        self.results.units = Hash(
            {"bins": ureg.angstrom, "number_densities": ureg.angstrom**-3}
        )

        # Preallocate dictionary to hold charge density profiles, if
        # necessary
        if self._charges is not None:
            self.results.charge_densities = Hash()
            self.results.units["charge_densities"] = (
                ureg.elementary_charge / ureg.angstrom**3
            )

        # Store time information, if necessary
        if not self._average:
            self.results.times = np.fromiter(
                (ts.time for ts in self._sliced_trajectory),
                dtype=float,
                count=self.n_frames,
            )
            self.results.units["times"] = ureg.picosecond

    def _single_frame(self) -> None:

        # Store atom or center-of-mass positions in the current frame
        for ag, gr, s in zip(self._groups, self._groupings, self._slices):
            self._positions[s] = (
                ag.positions if gr == "atoms" else center_of_mass(ag, gr)
            )

        if self._recenter is not None:

            # Globally unwrap entity positions for correct centers of
            # mass
            unwrap(
                self._positions,
                self._positions_old,
                self._dimensions,
                thresholds=self._thresholds,
                images=self._images,
            )

            # Shift entity positions by the difference between the group
            # and target centers of mass
            self._positions -= np.fromiter(
                (
                    0 if np.isnan(tx) else gx - tx
                    for gx, tx in zip(
                        center_of_mass(
                            positions=self._positions[self._recenter[0]],
                            masses=self._masses[self._recenter[0]],
                        ),
                        self._recenter[1],
                    )
                ),
                dtype=float,
                count=3,
            )

        # Wrap entity positions back into the simulation box so that
        # they belong to a histogram bin
        wrap(self._positions, self._dimensions)

        # Compute and tally the bin counts for the entity positions
        for ax, ia, n in zip(self.axes, self._axis_indices, self._n_bins):
            for ig, (gr, s) in enumerate(zip(self._groupings, self._slices)):
                if self._average:
                    self.results.number_densities[ax][ig] += numba_histogram(
                        self._positions[s, ia], n, self.results.bin_edges[ax]
                    )
                else:
                    self.results.number_densities[ax][ig, self._frame_index] = (
                        numba_histogram(
                            self._positions[s, ia], n, self.results.bin_edges[ax]
                        )
                    )

    def _single_frame_parallel(
        self, index: int
    ) -> tuple[int, dict[str, np.ndarray[float]]]:

        # Set current trajectory frame
        self._sliced_trajectory[index]

        # Preallocate arrays to hold bin counts for the current frame
        counts = {
            ax: np.empty((self._n_groups, n)) for ax, n in zip(self.axes, self._n_bins)
        }

        # Calculate or get entity positions
        if self._recenter is None:
            positions = np.empty((self._N, 3))
            for ag, gr, s in zip(self._groups, self._groupings, self._slices):
                positions[s] = ag.positions if gr == "atoms" else center_of_mass(ag, gr)
            wrap(positions, self._dimensions)
        else:
            positions = self._positions[index]

        # Compute and tally the bin counts for the entity positions
        for ax, ia, n in zip(self.axes, self._axis_indices, self._n_bins):
            for ig, (gr, s) in enumerate(zip(self._groupings, self._slices)):
                counts[ax][ig] = numba_histogram(
                    positions[s, ia], n, self.results.bin_edges[ax]
                )

        return index, counts

    def _conclude(self) -> None:

        # Consolidate parallel results and clean up memory by deleting
        # arrays that will not be reused
        if self._parallel:
            self._results = sorted(self._results)
            self.results.number_densities = Hash()
            for ax in self.axes:
                self.results.number_densities[ax] = np.stack(
                    [r[1][ax] for r in self._results], axis=1
                )
                if self._average:
                    self.results.number_densities[ax] = self.results.number_densities[
                        ax
                    ].sum(axis=1)

            del self._results
            if self._recenter is not None:
                del self._positions
        else:
            del self._positions
        if self._recenter is not None:
            del self._positions_old, self._images, self._thresholds

        # Normalize histograms by bin volume
        volume = np.prod(self._dimensions)
        for ax, n in zip(self.axes, self._n_bins):
            denom = volume / n
            if self._average:
                denom *= self.n_frames
            self.results.number_densities[ax] /= denom

            if self._charges is not None:
                self.results.charge_densities[ax] = np.einsum(
                    "g,g...b->...b", self._charges, self.results.number_densities[ax]
                )

    def _validate_input(
        self, value: Any, unit_: str, name: str, n_axes: int
    ) -> Union[Real, np.ndarray[Real]]:
        """
        Validates and processes input values to the calculation methods.

        Parameters
        ----------
        value : `Any`
            Value to validate and process.

        unit_ : `str`
            Reference unit for the value.

        name : `str`
            Name of the value being validated.

        n_axes : `int`
            Number of axes along which to calculate the value.

        Returns
        -------
        `Real` or `numpy.ndarray`
            Processed value.
        """

        if value is None:
            value = n_axes * [value]
        else:
            if self._reduced and not is_unitless(value):
                emsg = f"'{name}' cannot have units when 'reduced=True'."
                raise TypeError(emsg)
            value = strip_unit(value, unit_)[0]
            if isinstance(value, Real):
                value = n_axes * [value]
            elif len(value) != n_axes:
                emsg = f"The length of '{name}' must match the " "number of axes."
                raise ValueError(emsg)
        return value

    def calculate_surface_charge_densities(
        self,
        axes: Union[str, tuple[str]] = None,
        dielectrics: Union[float, tuple[float]] = None,
        *,
        dVs: Union[float, np.ndarray[float], "unit.Quantity", Q_] = None,
    ) -> None:
        """
        Calculates the surface charge densities :math:`\\sigma_q` for
        the specified system dimensions using the charge density
        profiles :math:`\\rho_q(z)`.

        Parameters
        ----------
        axes : `str` or array-like, optional
            Axes along which to compute the potential profiles. If not
            specified, all axes for which charge density profiles were
            calculated will be used.

            **Examples**::code:`"xy"` or :code:`("x", "y")`
            for the :math:`x`- and :math:`y`-directions.

        dielectrics : `float`, optional
            Relative permittivities or dielectric constants
            :math:`\\varepsilon_\\mathrm{r}`. Only optional if
            previously provided to another calculation method in this
            class.

        dVs : `float`, array-like, `openmm.unit.Quantity`, or \
        `pint.Quantity`, keyword-only, optional
            Potential differences :math:`\\Delta\\Psi` across the system
            dimensions specified in `axes`. Can be retrieved if
            previously provided to another calculation method in this
            class.

            **Shapes**: :math:`(N_\\mathrm{axes},)` or
            :math:`(N_\\mathrm{axes},\\,N_\\mathrm{frames})`.

            **Reference unit**: :math:`\\mathrm{V}`.
        """

        # Ensure charge density profiles have already been calculated
        if not hasattr(self.results, "charge_densities"):
            emsg = (
                "Either call DensityProfile.run() before "
                "DensityProfile.calculate_potential_profiles() or "
                "provide charge information when initializing the "
                "DensityProfile object."
            )
            raise RuntimeError(emsg)

        # Validate inputs
        if axes is None:
            axes = self.axes
            axis_indices = self._axis_indices
        else:
            if isinstance(axes, Real) or any(not isinstance(ax, str) for ax in axes):
                raise ValueError("'axes' must only contain strings.")
            axes = tuple(axes)
            axis_indices = [ord(ax.lower()) - 120 for ax in axes]
        try:
            relative_axes = [self.axes.index(ax) for ax in axes]
        except ValueError:
            raise ValueError("Invalid axis passed in 'axes'.")
        n_axes = len(axes)
        dimensions = self._dimensions[axis_indices]

        if dielectrics is not None:
            self._dielectrics[relative_axes] = self._validate_input(
                dielectrics, "", "dielectrics", n_axes
            )
        elif np.any(np.isnan(self._dielectrics)):
            raise ValueError("No dielectric constants found or provided.")

        if dVs is not None:
            self._dVs[relative_axes] = self._validate_input(dVs, "V", "dVs", n_axes)

        # Preallocate dictionary to hold potential profiles
        if not hasattr(self.results, "surface_charge_densities"):
            shape = [self._n_axes]
            if not self._average:
                shape.append(self.n_frames)
            self.results.surface_charge_densities = np.empty(shape)
            self.results.surface_charge_densities[:] = np.nan
            self.results.units["surface_charge_densities"] = (
                ureg.elementary_charge / ureg.angstrom**2
            )

        # Calculate surface charge densities
        for i, (ax, dielectric, L, dV) in enumerate(
            zip(axes, self._dielectrics, dimensions, self._dVs)
        ):
            self.results.surface_charge_densities[i] = calculate_surface_charge_density(
                self.results.bins[ax],
                self.results.charge_densities[ax],
                dielectric,
                L=L,
                dV=None if np.isnan(dV) else dV,
                reduced=self._reduced,
            )

    def calculate_potential_profiles(
        self,
        axes: Union[str, tuple[str]] = None,
        dielectrics: Union[float, tuple[float]] = None,
        *,
        sigmas_q: Union[float, np.ndarray[float], "unit.Quantity", Q_] = None,
        dVs: Union[float, np.ndarray[float], "unit.Quantity", Q_] = None,
        thresholds: Union[float, np.ndarray[float]] = 1e-5,
        V0s: Union[float, np.ndarray[float], "unit.Quantity", Q_] = 0,
        methods: Union[str, tuple[str]] = "integral",
        pbcs: Union[bool, tuple[bool]] = False,
    ) -> None:
        """
        Calculates the potential profiles in the specified dimensions
        using the charge density profiles by numerically solving
        Poisson's equation for electrostatics.

        Parameters
        ----------
        axes : `str` or array-like, optional
            Axes along which to compute the potential profiles. If not
            specified, all axes for which charge density profiles were
            calculated will be used.

            **Examples**::code:`"xy"` or :code:`("x", "y")`
            for the :math:`x`- and :math:`y`-directions.

        dielectrics : `float`, optional
            Relative permittivities or dielectric constants
            :math:`\\varepsilon_\\mathrm{r}`. Only optional if
            previously provided to another calculation method in this
            class.

        sigmas_q : `float`, array-like, `openmm.unit.Quantity`, or \
        `pint.Quantity`, keyword-only, optional
            Surface charge densities :math:`\\sigma_q`. Used to ensure
            that the electric field in the bulk of the solution is zero.
            If not provided, it is determined using `dVs` and the charge
            density profiles, or the average values in the centers of
            the integrated charge density profiles.

            .. note::

               :math:`\\sigma_q` and :math:`\\Delta\\Psi` should have
               the same sign.

            **Shapes**: :math:`(N_\\mathrm{axes},)` or
            :math:`(N_\\mathrm{axes},\\,N_\\mathrm{frames})`.

            **Reference unit**: :math:`\\mathrm{e/Å^2}`.

        dVs : `float`, array-like, `openmm.unit.Quantity`, or \
        `pint.Quantity`, keyword-only, optional
            Potential differences :math:`\\Delta\\Psi` across the system
            dimensions specified in `axes`. Can be retrieved if
            previously provided to another calculation method in this
            class. Has no effect if `sigmas_q` is provided since this
            value is used solely to calculate `sigmas_q`.

            **Shapes**: :math:`(N_\\mathrm{axes},)` or
            :math:`(N_\\mathrm{axes},\\,N_\\mathrm{frames})`.

            **Reference unit**: :math:`\\mathrm{V}`.

        thresholds : `float` or array-like, keyword-only, \
        default: :code:`1e-5`
            Thresholds for determining the plateau regions of the
            integrals of the charge density profiles to calculate
            `sigmas_q`. Has no effect if `sigmas_q` is provided, or if
            `sigmas_q` can be calculated using `dVs` and
            `charge_density_profiles`.

        V0s : `float`, array-like, `openmm.unit.Quantity`, or \
        `pint.Quantity`, keyword-only, default: :code:`0`
            Potentials :math:`\\Psi_0` at the left boundary.

            **Shapes**: :math:`(N_\\mathrm{axes},)` or
            :math:`(N_\\mathrm{axes},\\,N_\\mathrm{frames})`.

            **Reference unit**: :math:`\\mathrm{V}`.

        methods : `str` or array-like, keyword-only, \
        default: :code:`"integral"`
            Methods to use to calculate the potential profiles.

            **Valid values**: :code:`"integral"`, :code:`"matrix"`.

        pbcs : `bool`, keyword-only, default: :code:`False`
            Specifies whether the system has periodic boundary
            conditions in each of the axes. Only used when
            :code:`method="matrix"`.
        """

        # Ensure charge density profiles have already been calculated
        if not hasattr(self.results, "charge_densities"):
            emsg = (
                "Either call DensityProfile.run() before "
                "DensityProfile.calculate_potential_profiles() or "
                "provide charge information when initializing the "
                "DensityProfile object."
            )
            raise RuntimeError(emsg)

        # Preallocate dictionary to hold potential profiles
        if not hasattr(self.results, "potentials"):
            self.results.potentials = Hash()
            self.results.units["potentials"] = ureg.volt

        # Validate inputs
        if axes is None:
            axes = self.axes
            axis_indices = self._axis_indices
        else:
            if isinstance(axes, Real) or any(not isinstance(ax, str) for ax in axes):
                raise ValueError("'axes' must only contain strings.")
            axes = tuple(axes)
            axis_indices = [ord(ax.lower()) - 120 for ax in axes]
        try:
            relative_axes = [self.axes.index(ax) for ax in axes]
        except ValueError:
            raise ValueError("Invalid axis passed in 'axes'.")
        n_axes = len(axes)
        dimensions = self._dimensions[axis_indices]

        if dielectrics is not None:
            self._dielectrics[relative_axes] = self._validate_input(
                dielectrics, "", "dielectrics", n_axes
            )
        elif np.any(np.isnan(self._dielectrics)):
            raise ValueError("No dielectric constants found or provided.")

        if sigmas_q is None:
            if hasattr(self.results, "surface_charge_densities") and np.all(
                np.isfinite(self._dVs[relative_axes])
            ):
                sigmas_q = -self.results.surface_charge_densities
            else:
                sigmas_q = self._validate_input(sigmas_q, "e/Å^2", "sigmas_q", n_axes)
                if dVs is not None:
                    self._dVs[relative_axes] = self._validate_input(
                        dVs, "V", "dVs", n_axes
                    )
        else:
            sigmas_q = self._validate_input(sigmas_q, "e/Å^2", "sigmas_q", n_axes)

        V0s = self._validate_input(V0s, "V", "V0s", n_axes)
        thresholds = self._validate_input(thresholds, "", "thresholds", n_axes)

        METHODS = {"integral", "matrix"}
        if isinstance(methods, str) and methods in METHODS:
            methods = n_axes * [methods]
        elif len(methods) != n_axes:
            emsg = "The length of 'methods' must match the number of axes."
            raise ValueError(emsg)
        else:
            for method in methods:
                if method not in METHODS:
                    emsg = (
                        f"Invalid method '{method}'. Valid values: "
                        "'" + "', '".join(METHODS) + "'."
                    )
                    raise ValueError(emsg)

        if isinstance(pbcs, bool):
            pbcs = n_axes * [pbcs]
        elif len(pbcs) != n_axes:
            emsg = "The length of 'pbcs' must match the number of axes."
            raise ValueError(emsg)
        elif not all(isinstance(pbc, bool) for pbc in pbcs):
            raise ValueError("All values in 'pbcs' must be booleans.")

        # Calculate potential profiles
        for ax, dielectric, L, sigma_q, dV, threshold, V0, method, pbc in zip(
            axes,
            self._dielectrics,
            dimensions,
            sigmas_q,
            self._dVs,
            thresholds,
            V0s,
            methods,
            pbcs,
        ):
            self.results.potentials[ax] = calculate_potential_profile(
                self.results.bins[ax],
                self.results.charge_densities[ax],
                dielectric,
                L=L,
                sigma_q=sigma_q,
                dV=None if np.isnan(dV) else dV,
                threshold=threshold,
                V0=V0,
                method=method,
                pbc=pbc,
                reduced=self._reduced,
            )



"""
Algorithms
==========
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module is a collection of algorithms used in other MDCraft
(sub)modules.
"""

from . import accelerated, correlation, molecule, topology, unit, utility

__all__ = ["accelerated", "correlation", "molecule", "topology", "unit", "utility"]



"""
OpenMM topology transformations
===============================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains implementations of common OpenMM topology
transformations, like the generation of initial particle positions and
subsetting existing topology objects.
"""

from itertools import repeat
from typing import Any, Iterable, Union

import numpy as np
from openmm import app

from ..algorithm import topology as t


def create_atoms(*args, **kwargs) -> Any:
    """
    Generates initial particle positions.

    This is an alias function. For more information, see
    :func:`mdcraft.algorithm.topology.create_atoms`.
    """

    return t.create_atoms(*args, **kwargs)


def _get_hierarchy_indices(
    item: Union[app.Atom, app.topology.Bond, app.Residue, app.Chain],
    bonds: dict[str, list],
) -> tuple[set[int], set[int], set[int], set[int]]:
    """
    Get unique indices of all topology items related to the one passed
    to this function.

    For an atom, the indices of itself, the residue it belongs to, and
    the chain that its residue belongs to are returned.

    For a bond, the indices of the two atoms it connects, itself, the
    residue(s) that the two atoms belong to, and the chain(s) that the
    residue(s) belong to are returned.

    For a residue, the indices of all underlying atoms and bonds,
    itself, and the chain it belongs to are returned.

    For a chain, the indices of all underlying atoms, bonds, and
    residues, and itself are returned.

    Parameters
    ----------
    item : `openmm.app.Atom`, `openmm.app.topology.Bond`,
    `openmm.app.Residue`, or `openmm.app.Chain`
        Topology item of interest.

    Returns
    -------
    atoms : `set`
        Indices of all atoms associated with `item`.

    bonds : `set`
        Indices of all bonds associated with `item`.

    residues : `set`
        Indices of all residues associated with `item`.

    chains : `set`
        Indices of all chains associated with `item`.
    """

    if isinstance(item, app.Atom):
        return {item.index}, set(), {item.residue.index}, {item.residue.chain.index}

    elif isinstance(item, app.topology.Bond):
        return (
            {item.atom1.index, item.atom2.index},
            {bonds.index(item)},
            {item.atom1.residue.index, item.atom2.residue.index},
            {item.atom1.residue.chain.index, item.atom2.residue.chain.index},
        )

    elif isinstance(item, app.Residue):
        return (
            {a.index for a in item.atoms()},
            {bonds.index(b) for b in item.bonds()},
            {item.index},
            {item.chain.index},
        )

    elif isinstance(item, app.Chain):
        atom_indices = set()
        bond_indices = set()
        residue_indices = set()
        for residue in item.residues():
            a, b, r, _ = _get_hierarchy_indices(residue, bonds)
            atom_indices |= a
            bond_indices |= b
            residue_indices |= r
        return atom_indices, bond_indices, residue_indices, {item.index}


def _is_topology_object(obj: Any):
    """
    Check if the argument is a topology item.

    Parameters
    ----------
    obj : `Any`
        Any object.

    Returns
    -------
    is_topology_object : `bool`
        Boolean value indicating whether `obj` is a topology item.
    """
    return isinstance(obj, (app.Atom, app.topology.Bond, app.Residue, app.Chain))


def get_subset(
    topology: app.Topology,
    positions: np.ndarray[float],
    *,
    delete: list[Any] = None,
    keep: list[Any] = None,
    types: Union[str, Iterable[str]] = None,
) -> Union[app.Topology, np.ndarray[float]]:
    r"""
    Creates a topology subset and gets its corresponding particle positions.

    Parameters
    ----------
    topology : `openmm.app.Topology`
        OpenMM topology.

    positions : `numpy.ndarray`
        Positions of the :math:`N` particles in the topology.

        **Shape**: :math:`(N,\,3)`.

    delete : array-like, keyword-only, optional
        `openmm.app.Atom`, `openmm.app.Bond`, `openmm.app.Residue`,
        and/or `openmm.app.Chain` objects, or the indices of those
        objects. If indices are provided, their corresponding object
        types (:code:`"atom"`, :code:`"residue"`, :code:`"chain"`) must
        be provided in `types`. The specified items will be deleted from
        the model.

        .. note::

           Only one of `delete` and `keep` can be specified.

    keep : array-like, keyword-only, optional
        `openmm.app.Atom`, `openmm.app.Bond`, `openmm.app.Residue`,
        and/or `openmm.app.Chain` objects, or the indices of those
        objects. If indices are provided, their corresponding object
        types (:code:`"atom"`, :code:`"residue"`, :code:`"chain"`) must
        be provided in `types`. The specified items will be kept in the
        model.

        .. note::

           Only one of `delete` and `keep` can be specified.

    types : `str` or array-like, keyword-only, optional
        Object types corresponding to the indices provided in `delete`
        or `keep`. If a `str` is provided, all items in the array are
        assumed to have the same object type. Must be provided if not
        all items in `delete` or `keep` are OpenMM topology objects.

    Returns
    -------
    topology : `openmm.app.Topology`
        OpenMM topology subset.

    positions : `numpy.ndarray`
        Positions of the remaining :math:`N_\mathrm{rem}` particles in
        the topology.

        **Shape**: :math:`(N_\mathrm{rem},\,3)`.
    """

    # Set boolean flags for subroutines below
    found = (delete is not None, keep is not None)

    # Check if both delete and keep arguments were provided
    if all(found):
        emsg = (
            "Only specify topology items to either delete or keep. "
            "When both types are specified, the atoms, bonds, "
            "residues, and/or chains to be removed from the topology "
            "become ambiguous."
        )
        raise ValueError(emsg)

    # Return original topology and positions if no items are specified
    elif not any(found):
        return topology, positions

    # Ensure object type(s) are provided if not all items are topology
    # objects
    elif types is None and not all(
        _is_topology_object(i) for i in next(a for a in [delete, keep] if a is not None)
    ):
        emsg = (
            "Object types must be specified for the topology items "
            f"to be {'kept' if found[0] else 'deleted'}."
        )
        raise ValueError(emsg)

    # Create a generator of types if a string is provided
    elif isinstance(types, str):
        same = True
        types = repeat(types)
    elif types is not None:
        same = all(t == "atoms" for t in types)

    # Create OpenMM Modeller
    modeller = app.Modeller(topology, positions)

    if types is not None:

        # Create dictionary with topology subitems
        model = {
            "atom": list(topology.atoms()),
            "bond": list(topology.bonds()),
            "chain": list(topology.chains()),
            "residue": list(topology.residues()),
        }

        # If indices and types of objects to be deleted are specified,
        # create an iterable object of corresponding items from the
        # dictionary
        if found[0]:
            delete = (
                i if _is_topology_object(i) else model[t][i]
                for i, t in zip(delete, types)
            )
        else:

            # Preallocate sets to store indices of atoms, residues, and
            # chains to keep
            atoms = set()
            bonds = set()
            residues = set()
            chains = set()

            # Remove items to be kept from the master list of topology
            # subitems to delete
            for item, item_type in zip(keep, types):
                if not _is_topology_object(item):
                    item = model[item_type][item]
                a, b, r, c = _get_hierarchy_indices(item, model["bond"])
                atoms |= a
                bonds |= b
                residues |= r
                chains |= c

            model["atom"] = np.delete(model["atom"], list(atoms))
            model["residue"] = np.delete(model["residue"], list(residues))
            model["chain"] = np.delete(model["chain"], list(chains))
            if not bonds and same:
                model["bond"] = []
            else:
                for i in sorted(bonds, reverse=True):
                    del model["bond"][i]
            delete = [i for t in model.values() for i in t]

    # Create subset by deleting objects from original topology
    modeller.delete(delete)

    return modeller.topology, modeller.positions



"""
Unit manipulation
=================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains helper functions for unit manipulation and reduction.
"""

from numbers import Number
from typing import Any, Union

import numpy as np

from .. import FOUND_OPENMM, Q_, ureg

if FOUND_OPENMM:
    from openmm import unit
    from ..openmm.unit import VACUUM_PERMITTIVITY


def get_scale_factors(
    bases: dict[str, Union["unit.Quantity", Q_]], other: dict[str, list] = None
) -> dict[str, Union["unit.Quantity", Q_]]:
    r"""
    Evaluates scale factors for reduced units.

    Parameters
    ----------
    bases : `dict`
        Fundamental quantities: molar mass (:math:`m`), length
        (:math:`\sigma`), and energy (:math:`\epsilon`).

        .. container::

           **Format**:

           .. code::

              {
                "mass": <openmm.unit.Quantity> | <pint.Quantity>,
                "length": <openmm.unit.Quantity> | <pint.Quantity>,
                "energy": <openmm.unit.Quantity> | <pint.Quantity>
              }

        **Reference units**: :math:`\mathrm{g/mol}`, :math:`\mathrm{nm}`,
        and :math:`\mathrm{kJ}`.

    other : `dict`, optional
        Other scale factors to compute. The key should be the name of
        the scale factor, and the value should contain `tuple`
        objects with the names of bases or default scale factors and
        their powers.

        **Example**:
        :code:`{"diffusivity": (("length", 2), ("time", -1))}`.

    Returns
    -------
    scales : `dict`
        Scale factors.
    """

    if other is not None:
        for name, params in other.items():
            factor = 1
            for base, power in params:
                factor *= bases[base] ** power
            bases[name] = factor

    return bases


def get_lj_scale_factors(
    bases: dict[str, Union["unit.Quantity", Q_]], other: dict[str, list] = None
) -> dict[str, Union["unit.Quantity", Q_]]:
    r"""
    Evaluates scale factors for Lennard-Jones reduced units.

    By default, the following scale factors are calculated:

    ===========================  ===================================================
    Quantity (`dict` key)        Expression
    ===========================  ===================================================
    :code:`"molar_energy"`       :math:`N_\mathrm{A}\epsilon`
    :code:`"time"`               :math:`\sqrt{m\sigma^2/\epsilon}`
    :code:`"velocity"`           :math:`\sigma/\tau`
    :code:`"force"`              :math:`\epsilon/\sigma`
    :code:`"temperature"`        :math:`\epsilon/k_\mathrm{B}T`
    :code:`"pressure"`           :math:`\epsilon/\sigma^3`
    :code:`"dynamic_viscosity"`  :math:`\epsilon\tau/\sigma^3`
    :code:`"charge"`             :math:`\sqrt{4\pi\varepsilon_0\sigma\epsilon}`
    :code:`"dipole"`             :math:`\sqrt{4\pi\varepsilon_0\sigma^3\epsilon}`
    :code:`"electric_field"`     :math:`\sqrt{\epsilon/(4\pi\varepsilon_0\sigma^3)}`
    :code:`"mass_density"`       :math:`m/\sigma^3`
    ===========================  ===================================================

    Parameters
    ----------
    bases : `dict`
        Fundamental quantities: molar mass (:math:`m`), length
        (:math:`\sigma`), and energy (:math:`\epsilon`).

        .. container::

           **Format**:

           .. code::

              {
                "mass": <openmm.unit.Quantity> | <pint.Quantity>,
                "length": <openmm.unit.Quantity> | <pint.Quantity>,
                "energy": <openmm.unit.Quantity> | <pint.Quantity>
              }

        **Reference units**: :math:`\mathrm{g/mol}`, :math:`\mathrm{nm}`,
        and :math:`\mathrm{kJ}`.

    other : `dict`, optional
        Other scale factors to compute. The key should be the name of
        the scale factor, and the value should contain `tuple`
        objects with the names of bases or default scale factors and
        their powers.

        **Example**:
        :code:`{"diffusivity": (("length", 2), ("time", -1))}`.

    Returns
    -------
    scales : `dict`
        Scale factors.
    """

    if bases["mass"].__module__ == "pint":
        avogadro_constant = ureg.avogadro_constant
        boltzmann_constant = ureg.boltzmann_constant
        bases["molar_energy"] = bases["energy"] * avogadro_constant
        bases["time"] = np.sqrt(
            bases["mass"] * bases["length"] ** 2 / bases["molar_energy"]
        ).to(ureg.picosecond)
        bases["charge"] = np.sqrt(
            4 * np.pi * ureg.vacuum_permittivity * bases["length"] * bases["energy"]
        ).to(ureg.elementary_charge)
    else:
        avogadro_constant = unit.AVOGADRO_CONSTANT_NA
        boltzmann_constant = unit.BOLTZMANN_CONSTANT_kB
        bases["molar_energy"] = bases["energy"] * avogadro_constant
        bases["time"] = (
            (bases["mass"] * bases["length"] ** 2 / bases["molar_energy"])
            .sqrt()
            .in_units_of(unit.picosecond)
        )
        bases["charge"] = (
            (4 * np.pi * VACUUM_PERMITTIVITY * bases["length"] * bases["energy"])
            .sqrt()
            .in_units_of(unit.elementary_charge)
        )

    # Define the default scale factors
    bases["velocity"] = bases["length"] / bases["time"]
    bases["force"] = bases["molar_energy"] / bases["length"]
    bases["temperature"] = bases["energy"] / boltzmann_constant
    bases["pressure"] = bases["energy"] / bases["length"] ** 3
    bases["dynamic_viscosity"] = bases["pressure"] * bases["time"]
    bases["dipole"] = bases["length"] * bases["charge"]
    bases["electric_field"] = bases["force"] / bases["charge"]
    bases["mass_density"] = bases["mass"] / (bases["length"] ** 3 * avogadro_constant)

    return get_scale_factors(bases, other)


def strip_unit(
    value: Union[Number, str, "unit.Quantity", Q_],
    unit_: Union[str, "unit.Unit", ureg.Unit] = None,
) -> tuple[Number, Union[None, "unit.Unit", ureg.Unit]]:
    """
    Strips the unit from an :obj:`openmm.unit.quantity.Quantity` or
    :obj:`pint.Quantity` object.

    Parameters
    ----------
    value : `numbers.Number`, `str`, `openmm.unit.Quantity`, or \
    `pint.Quantity`
        Physical quantity for which to get the magnitude of in the
        unit specified in `unit_`.

    unit_ : `str`, `openmm.unit.Unit`, or `pint.Unit`, optional
        Unit to convert to. If not specified, the original unit is used.

    Returns
    -------
    value : `numbers.Number`
        Magnitude of the physical quantity in the specified unit.

    unit_ : `openmm.unit.Unit` or `pint.Unit`
        Unit of the physical quantity.

    Examples
    --------
    For any quantity other than a :obj:`openmm.unit.quantity.Quantity`
    or :obj:`pint.Quantity` object, the raw quantity and user-specified
    unit are returned.

    >>> strip_unit(90.0, "deg")
    (90.0, 'deg')
    >>> strip_unit(90.0, ureg.degree)
    (90.0, <Unit('degree')>)

    If no target unit is specified, the magnitude and original unit of
    the quantity are returned.

    >>> strip_unit(1.380649e-23)
    (1.380649e-23, None)
    >>> strip_unit(1.380649e-23 * ureg.joule * ureg.kelvin ** -1)
    (1.380649e-23, <Unit('joule / kelvin')>)

    If a target unit using the same module as the quantity is specified,
    the quantity is first converted to the target unit, if necessary,
    before its magnitude and unit are returned.

    >>> g = 9.80665 * ureg.meter / ureg.second ** 2
    >>> strip_unit(g, "meter/second**2")
    (9.80665, <Unit('meter / second ** 2')>)
    >>> strip_unit(g, ureg.foot / ureg.second ** 2)
    (32.17404855643044, <Unit('foot / second ** 2')>)

    If a target unit using a different module than the quantity is
    specified, the quantity is converted to the specified target unit in
    the new module, if necessary, before its magnitude and unit are
    returned.

    >>> strip_unit(8.205736608095969e-05 * unit.meter ** 3 * unit.atmosphere
    ...            / (unit.kelvin * unit.mole),
    ...            ureg.joule / (ureg.kelvin * ureg.mole))
    (8.31446261815324, <Unit('joule / kelvin / mole')>)
    >>> strip_unit(8.205736608095969e-05 * ureg.meter ** 3 * ureg.atmosphere
    ...            / (ureg.kelvin * ureg.mole),
    ...            unit.joule / (unit.kelvin * unit.mole))
    (8.31446261815324, Unit({BaseUnit(..., name="kelvin", ...): -1.0, ...}))

    This function also supports strings directly:

    >>> strip_unit("299792458 m/s")
    (299792458.0, 'meter / second')
    >>> strip_unit("299792458 m/s", "ft/s")
    (983571056.4304463, 'foot / second')
    >>> strip_unit("299792458 m/s", unit.foot / unit.second)
    (983571056.4304463, Unit({BaseUnit(..., name="foot", ...): 1.0, ...}))
    >>> strip_unit("299792458 m/s", ureg.foot / ureg.second)
    (983571056.4304463, <Unit('foot / second')>)
    """

    def convert_openmm_to_pint(ou: "unit.Unit") -> ureg.Unit:
        """
        Converts an OpenMM unit to a Pint unit.

        Parameters
        ----------
        ou : `openmm.unit.Unit`
            OpenMM unit to convert.

        Returns
        -------
        pu : `pint.Unit`
            Pint unit equivalent to the OpenMM unit.
        """

        pu = ureg.Unit("")
        for u, p in ou.iter_base_or_scaled_units():
            pu *= ureg.Unit(u.name.replace(" ", "_")) ** p
        return pu

    if isinstance(value, Q_):

        # No target unit (unit_) specified; return Pint magnitude and
        # unit (unit__)
        if unit_ is None:
            unit__ = value.units
            value = value.magnitude
        else:

            # Convert OpenMM target unit (unit__ = unit_) to Pint unit
            # (unit_) and return unit__
            if getattr(unit_, "__module__", None) == "openmm.unit.unit":
                unit_, unit__ = convert_openmm_to_pint(unit_), unit_

            # Convert str or Pint target unit (unit_) to Pint unit
            # (unit__)
            else:
                unit__ = ureg.Unit(unit_)

            # Get magnitude of Pint quantity in Pint unit (unit_)
            value = value.m_as(unit_)

    elif getattr(value, "__module__", None) == "openmm.unit.quantity":
        swap = False

        # No target unit (unit_) specified; return OpenMM magnitude and
        # unit (unit__)
        if unit_ is None:
            unit_ = unit__ = value.unit

        # Store OpenMM target unit (unit_) in return value (unit__)
        elif getattr(unit_, "__module__", None) == "openmm.unit.unit":
            unit__ = unit_
        else:

            # Determine whether unit_ and unit__ need to be swapped at
            # the end; str target unit should give OpenMM unit, while
            # Pint target unit should give Pint unit
            swap = not isinstance(unit_, str)

            # Convert str or Pint target unit (unit_) to OpenMM unit
            # (unit__)
            unit_ = ureg.Unit(unit_)
            try:
                unit__ = unit.dimensionless
                for u, p in unit_._units.items():
                    unit__ *= getattr(unit, u) ** p
            except AttributeError:
                emsg = (
                    "strip_unit() relies on the pint module for "
                    "parsing units. At least one unit in 'unit' is "
                    "not defined the same way in openmm.unit and pint, "
                    "so the unit conversion cannot be performed. Try "
                    "specifying a openmm.unit.Quantity object for "
                    "'unit' instead."
                )
                raise ValueError(emsg)
        value = value.value_in_unit(unit__)
        if swap:
            unit_, unit__ = unit__, unit_

    elif isinstance(value, str):
        value = ureg(value)
        if isinstance(value, Number):
            return value, unit_
        else:
            if unit_ is not None:
                unit__ = unit_
                if getattr(unit_, "__module__", None) == "openmm.unit.unit":
                    unit_ = convert_openmm_to_pint(unit_)
                value = value.to(unit_)
                if isinstance(unit__, str):
                    unit__ = str(value.units)
                value = value.magnitude
            else:
                value, unit__ = value.magnitude, str(value.units)

    else:
        unit__ = unit_

    return value, unit__


def is_unitless(value: Any) -> bool:
    """
    Determines whether a value is unitless.

    Parameters
    ----------
    value : `Any`
        Value to check for unitlessness.

    Returns
    -------
    is_unitless : `bool`
        Whether the value is unitless.

    Examples
    --------
    >>> is_unitless(90.0)
    True
    >>> is_unitless("90 degrees")
    False
    >>> is_unitless(90.0 * ureg.degree)
    False
    >>> is_unitless(90.0 * unit.degree)
    False
    >>> is_unitless({"quantity": 90 * ureg.degree})
    True
    """

    return strip_unit(value)[1] is None



"""
Correlation functions
=====================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains algorithms for evaluating the spatial and temporal
correlation between data sets, such as the autocorrelation and
cross-correlation functions and the closely related mean squared
displacement.
"""

from typing import Union
import warnings

import numpy as np
from scipy import fft


def correlation_fft(
    x: np.ndarray[Union[float, complex]],
    y: np.ndarray[Union[float, complex]] = None,
    /,
    axis: int = None,
    *,
    average: bool = False,
    double: bool = False,
    vector: bool = False,
) -> np.ndarray[Union[float, complex]]:
    r"""
    Evaluates the autocorrelation functions (ACF)
    :math:`\mathrm{R_\mathbf{XX}}(\tau)` or cross-correlation functions
    (CCF) :math:`\mathrm{R_\mathbf{XY}}(\tau)` of time series
    :math:`\mathbf{X}(t)` and :math:`\mathbf{Y}(t)` using fast Fourier
    transforms (FFT).

    The fast convolution algorithm (FCA) [1]_ [2]_ is associated wtih
    the Wiener–Khinchin theorem and has a time complexity of
    :math:`\mathcal{O}(N\log{N})`.

    The ACF for a data set :math:`\mathbf{X}(t)` can be computed using

    .. math::

       \begin{gather*}
         \hat{\mathbf{X}}(\omega)=\mathcal{F}[\mathbf{X}(t)]\\
         \mathrm{R}_{\mathbf{XX}}(\tau)=\mathcal{F}^{-1}
         [\hat{\mathbf{X}}(\omega)\hat{\mathbf{X}}^*(\omega)]
       \end{gather*}

    where :math:`\tau` is the time lag and the asterisk (:math:`^*`)
    denotes the complex conjugate.

    Similarly, the CCF for data sets :math:`\mathbf{X}(t)` and
    :math:`\mathbf{Y}(t)` can be computed using

    .. math::

       \mathrm{R}_{\mathbf{XY}}(\tau)=\mathcal{F}^{-1}[\mathcal{F}
       [\mathbf{X}(t)]\cdot\mathcal{F}[\mathbf{Y}(t)]]

    Parameters
    ----------
    x : `numpy.ndarray`, positional-only
        Time evolution of :math:`d`-dimensional data for :math:`N`
        entities over :math:`N_\mathrm{b}` blocks of :math:`N_t` times
        each.

        .. container::

           **Shape**:

           * Scalar data: :math:`(N_t,)`, :math:`(N_t,\,N)`,
             :math:`(N_\mathrm{b},\,N_t)`, or
             :math:`(N_\mathrm{b},\,N_t,\,N)`.
           * Vector data: :math:`(N_t,\,d)`, :math:`(N_t,\,N,\,d)`,
             :math:`(N_\mathrm{b},\,N_t,\,d)`, or
             :math:`(N_\mathrm{b},\,N_t,\,N,\,d)`.

    y : `numpy.ndarray`, positional-only, optional
        Time evolution of :math:`d`-dimensional data for another
        :math:`N` entities over :math:`N_\mathrm{b}` blocks of
        :math:`N_t` times each. If provided, the CCF for `x` and `y` is
        calculated. Otherwise, the ACF for `x` is calculated.

        **Shape**: Same as `x`.

    axis : `int`, optional
        Axis along which time evolves. If not specified, the axis is
        determined automatically using the shape of `x`.

    average : `bool`, keyword-only, default: :code:`True`
        Determines whether the ACF/CCF is averaged over all entities.
        Only available if `x` and `y` contain information for multiple
        entities.

    double : `bool`, keyword-only, default: :code:`False`
        Determines whether the ACF is doubled or the negative and
        positive time lags are combined for the CCF.

    vector : `bool`, keyword-only, default: :code:`False`
        Specifies whether `x` and `y` contain vectors. If :code:`True`,
        the ACF/CCF is summed over the last axis.

    Returns
    -------
    corr : `numpy.ndarray`
        ACF or CCF.

        .. container::

           **Shape**:

           For ACF, the shape is that of `x` but with the following
           modifications:

           * If :code:`average=True`, the axis indexing the :math:`N`
             entities is removed.
           * If :code:`vector=True`, the last axis is removed.

           For CCF, the shape is that of `x` but with the following
           modifications:

           * If :code:`average=True`, the axis indexing the :math:`N`
             entities is removed.
           * If :code:`double=False`, the axis indexing the
             :math:`N_t` times now has a length of :math:`2N_t-1` to
             accomodate negative and positive time lags.
           * If :code:`vector=True`, the last axis is removed.

    References
    ----------
    .. [1] Kneller, G. R.; Keiner, V.; Kneller, M.; Schiller, M.
       NMOLDYN: A Program Package for a Neutron Scattering Oriented
       Analysis of Molecular Dynamics Simulations. *Computer Physics
       Communications* **1995**, *91* (1–3), 191–214.
       https://doi.org/10.1016/0010-4655(95)00048-K.

    .. [2] Calandrini, V.; Pellegrini, E.; Calligari, P.; Hinsen, K.;
       Kneller, G. R. NMoldyn - Interfacing Spectroscopic Experiments,
       Molecular Dynamics Simulations and Models for Time Correlation
       Functions. *JDN* **2011**, *12*, 201–232.
       https://doi.org/10.1051/sfn/201112010.
    """

    # Ensure arrays have valid shapes
    x = np.asarray(x)
    if x.size == 0:
        raise ValueError("The arrays cannot be empty.")
    ndim = x.ndim
    if not 1 <= ndim <= 4:
        emsg = "The arrays must be one-, two-, three-, or four-" "dimensional."
        raise ValueError(emsg)
    if y is not None:
        y = np.asarray(y)
        if x.shape != y.shape:
            raise ValueError("The arrays must have the same shape.")

    # Check or set axis along which to compute the ACF/CCF
    if axis is None:
        if ndim == 4:
            axis = 1
        else:
            axis = 0
            if ndim > 1:
                wmsg = (
                    "The axis along which to compute the ACF/CCF "
                    "was not specified and is ambiguous for a "
                    "multidimensional array. As such, it has been "
                    "set to the first axis by default."
                )
                warnings.warn(wmsg)
    elif axis not in {0, 1}:
        emsg = "The ACF/CCF can only be computed along the first or " "second axis."
        raise ValueError(emsg)

    # Determine whether faster real-valued FFTs can be used
    if real := np.isrealobj(x) and (y is None or np.isrealobj(y)):
        fft_ = fft.rfft
        ifft = fft.irfft
    else:
        fft_ = fft.fft
        ifft = fft.ifft

    # Compute the power spectral density by first zero-padding the
    # arrays for linear convolution and then inverting it to get the
    # ACF/CCF
    N_t = x.shape[axis]
    n = 2 * fft.next_fast_len(N_t, real=real)
    if y is None:
        f = fft_(x, n=n, axis=axis)
        r = ifft(f * f.conj(), axis=axis)
        r = (double + 1) * (r[:, :N_t] if axis else r[:N_t])
    else:
        fx = fft_(x, n=n, axis=axis)
        fy = fft_(y, n=n, axis=axis)
        f = fx.conj() * fy
        if double:
            r = ifft(f + fx * fy.conj(), axis=axis)
            r = r[:, :N_t] if axis else r[:N_t]
        else:
            r = ifft(f, axis=axis)

    # Sum over the last axis if the arrays contain vectors
    if vector:
        r = r.sum(axis=-1)

    # Determine the axes over which to expand the dimensions of the
    # reversed time array for correct matrix division
    axes = list(range(ndim - vector))
    axes.remove(axis)

    # Normalize the ACF/CCF
    if axis:
        r[:, :N_t] /= np.expand_dims(np.arange(N_t, 0, -1), axes)
        if r.shape[axis] != N_t:
            r[:, 1 - N_t :] /= np.expand_dims(np.arange(1, N_t), axes)
            r = np.hstack((r[:, 1 - N_t :], r[:, :N_t]))
    else:
        r[:N_t] /= np.expand_dims(np.arange(N_t, 0, -1), axes)
        if r.shape[axis] != N_t:
            r[1 - N_t :] /= np.expand_dims(np.arange(1, N_t), axes)
            r = np.concatenate((r[1 - N_t :], r[:N_t]))

    # Average over all entities, if desired
    if average:
        axis_avg = ndim - vector - 1
        if axis != axis_avg:
            return r.mean(axis=axis_avg)

    return r


def correlation_shift(
    x: np.ndarray[Union[float, complex]],
    y: np.ndarray[Union[float, complex]] = None,
    /,
    axis: int = None,
    *,
    average: bool = False,
    double: bool = False,
    vector: bool = False,
) -> np.ndarray[Union[float, complex]]:
    r"""
    Evaluates the autocorrelation functions (ACF)
    :math:`\mathrm{R_\mathbf{XX}}(\tau)` or cross-correlation functions
    (CCF) :math:`\mathrm{R_\mathbf{XY}}(\tau)` of time series
    :math:`\mathbf{X}(t)` and :math:`\mathbf{Y}(t)` directly by using
    sliding windows.

    The ACF for a data set :math:`\mathbf{X}(t)` can be computed using

    .. math::

       \mathrm{R}_{\mathbf{XX}}(\tau)
       =\langle\mathbf{X}(t_0+\tau)\cdot\mathbf{X}^*(t_0)\rangle
       =\dfrac{1}{N_\tau}\sum_{j=1}^{N_\tau}
       \textbf{X}(t_j+\tau)\cdot\textbf{X}^*(t_j)

    where :math:`\tau` is the time lag, :math:`t_j` is an arbitrary
    reference time, :math:`N_\tau` is the number of possible reference
    times, and the asterisk (:math:`^*`) denotes the complex conjugate.

    Similarly, the CCF for data sets :math:`\mathbf{X}(t)` and
    :math:`\mathbf{Y}(t)` can be computed using

    .. math::

       \mathrm{R}_{\mathbf{XY}}(\tau)
       =\langle\mathbf{X}(t_0+\tau)\cdot\mathbf{Y}^*(t_0)\rangle
       =\dfrac{1}{N_\tau}\sum_{j=1}^{N_\tau}
       \textbf{X}(t_j+\tau)\cdot\textbf{Y}^*(t_j)

    To minimize statistical noise, the ACF/CCF is calculated for and
    averaged over  all possible reference times :math:`t_0`. As such,
    this algorithm has a time complexity of :math:`\mathcal{O}(N^2)`.
    With large data sets, this approach is too slow to be useful. If
    your machine supports fast Fourier transforms (FFT), use the much
    more performant FFT-based algorithm implemented in
    :func:`mdcraft.algorithm.correlation.correlation_fft` instead.

    Parameters
    ----------
    x : `numpy.ndarray`, positional-only
        Time evolution of :math:`d`-dimensional data for :math:`N`
        entities over :math:`N_\mathrm{b}` blocks of :math:`N_t` times
        each.

        .. container::

           **Shape**:

           * Scalar data: :math:`(N_t,)`, :math:`(N_t,\,N)`,
             :math:`(N_\mathrm{b},\,N_t)`, or
             :math:`(N_\mathrm{b},\,N_t,\,N)`.
           * Vector data: :math:`(N_t,\,d)`, :math:`(N_t,\,N,\,d)`,
             :math:`(N_\mathrm{b},\,N_t,\,d)`, or
             :math:`(N_\mathrm{b},\,N_t,\,N,\,d)`.

    y : `numpy.ndarray`, positional-only, optional
        Time evolution of :math:`d`-dimensional data for another
        :math:`N` entities over :math:`N_\mathrm{b}` blocks of
        :math:`N_t` times each. If provided, the CCF for `x` and `y` is
        calculated. Otherwise, the ACF for `x` is calculated.

        **Shape**: Same as `x`.

    axis : `int`, optional
        Axis along which time evolves. If not specified, the axis is
        determined automatically using the shape of `x`.

    average : `bool`, keyword-only, default: :code:`True`
        Determines whether the ACF/CCF is averaged over all entities.
        Only available if `x` and `y` contain information for multiple
        entities.

    double : `bool`, keyword-only, default: :code:`False`
        Determines whether the ACF is doubled or the negative and
        positive time lags are combined for the CCF.

    vector : `bool`, keyword-only, default: :code:`False`
        Specifies whether `x` and `y` contain vectors. If :code:`True`,
        the ACF/CCF is summed over the last axis.

    Returns
    -------
    corr : `numpy.ndarray`
        ACF or CCF.

        .. container::

           **Shape**:

           For ACF, the shape is that of `x` but with the following
           modifications:

           * If :code:`average=True`, the axis indexing the :math:`N`
             entities is removed.
           * If :code:`vector=True`, the last axis is removed.

           For CCF, the shape is that of `x` but with the following
           modifications:

           * If :code:`average=True`, the axis indexing the :math:`N`
             entities is removed.
           * If :code:`double=False`, the axis indexing the
             :math:`N_t` times now has a length of :math:`2N_t-1` to
             accomodate negative and positive time lags.
           * If :code:`vector=True`, the last axis is removed.
    """

    # Ensure arrays have valid shapes
    x = np.asarray(x)
    if x.size == 0:
        raise ValueError("The arrays cannot be empty.")
    ndim = x.ndim
    if not 1 <= ndim <= 4:
        emsg = "The arrays must be one-, two-, three-, or four-" "dimensional."
        raise ValueError(emsg)
    if y is not None:
        y = np.asarray(y)
        if x.shape != y.shape:
            raise ValueError("The arrays must have the same shape.")

    # Check or set axis along which to compute the ACF/CCF
    if axis is None:
        if ndim == 4:
            axis = 1
        else:
            axis = 0
            if ndim > 1:
                wmsg = (
                    "The axis along which to compute the ACF/CCF "
                    "was not specified and is ambiguous for a "
                    "multidimensional array. As such, it has been "
                    "set to the first axis by default."
                )
                warnings.warn(wmsg)
    elif axis not in {0, 1}:
        emsg = "The ACF/CCF can only be computed along the first or " "second axis."
        raise ValueError(emsg)

    # Calculate the ACF/CCF
    N_t = x.shape[axis]
    if y is None:
        if ndim == 1:
            r = np.fromiter(
                (np.dot(x[i:], x[: -i if i else None]) for i in range(N_t)),
                dtype=float,
                count=N_t,
            )
        elif axis:
            axes = f"bt...{'d' * vector}"
            r = np.stack(
                [
                    np.einsum(
                        f"{axes},{axes}->b...", x[:, i:], x[:, : -i if i else None]
                    )
                    for i in range(N_t)
                ],
                axis=1,
            )
        else:
            axes = f"t...{'d' * vector}"
            r = np.stack(
                [
                    np.einsum(f"{axes},{axes}->...", x[i:], x[: -i if i else None])
                    for i in range(N_t)
                ]
            )
    else:
        start = np.r_[np.zeros(N_t - 1, dtype=int), 0:N_t]
        stop = np.r_[1 : N_t + 1, N_t * np.ones(N_t - 1, dtype=int)]
        if ndim == 1:
            r = np.fromiter(
                (
                    np.dot(x[i:j], y[k:m])
                    for i, j, k, m in zip(start[::-1], stop[::-1], start, stop)
                ),
                dtype=float,
                count=2 * N_t - 1,
            )
        elif axis:
            axes = f"bt...{'d' * vector}"
            r = np.stack(
                [
                    np.einsum(f"{axes},{axes}->b...", x[:, i:j], y[:, k:m])
                    for i, j, k, m in zip(start[::-1], stop[::-1], start, stop)
                ],
                axis=1,
            )
        else:
            axes = f"t...{'d' * vector}"
            r = np.stack(
                [
                    np.einsum(f"{axes},{axes}->...", x[i:j], y[k:m])
                    for i, j, k, m in zip(start[::-1], stop[::-1], start, stop)
                ]
            )

    # Double the ACF or overlap the negative and positive time lags for
    # the CCF, if desired
    if double:
        if y is None:
            r *= 2
        elif axis:
            r = r[:, N_t - 1 :] + r[:, N_t - 1 :: -1]
        else:
            r = r[N_t - 1 :] + r[N_t - 1 :: -1]

    # Determine the axes over which to expand the dimensions of the
    # reversed time array for correct matrix division
    axes = list(range(ndim - vector))
    axes.remove(axis)

    # Normalize the ACF/CCF
    if axis:
        r[:, -N_t:] /= np.expand_dims(np.arange(N_t, 0, -1), axes)
        if r.shape[axis] != N_t:
            r[:, : N_t - 1] /= np.expand_dims(np.arange(1, N_t), axes)
    else:
        r[-N_t:] /= np.expand_dims(np.arange(N_t, 0, -1), axes)
        if r.shape[axis] != N_t:
            r[: N_t - 1] /= np.expand_dims(np.arange(1, N_t), axes)

    # Average over all entities, if desired
    if average:
        axis_avg = ndim - 1 - vector
        if axis != axis_avg:
            return r.mean(axis=axis_avg)

    return r


def msd_fft(
    r_i: np.ndarray[float],
    r_j: np.ndarray[float] = None,
    /,
    axis: int = None,
    *,
    average: bool = True,
) -> np.ndarray[float]:
    r"""
    Evaluates the mean squared displacements (MSD) or the analogous
    cross displacements (CD) of positions :math:`\mathbf{r}_i(t)` and
    :math:`\mathbf{r}_j(t)` using fast Fourier transforms (FFT).

    For a set of positions :math:`\mathbf{r}_i(t)`, the MSD is computed
    using the algorithm [1]_ [2]_

    .. math::

       \mathrm{MSD}_{i,\,m}&=\frac{1}{N_t-m}\sum_{k=0}^{N_t-m-1}
       [\textbf{r}_{i,\,k+m}-\textbf{r}_{i,\,k}]^2\\
       &=\frac{1}{N_t-m}\sum_{k=0}^{N_t-m-1}
       \left[\textbf{r}_{i,\,k+m}^2+\textbf{r}_{i,\,k}^2\right]
       -\frac{2}{N_t-m}\sum_{k=0}^{N_t-m-1}
       \textbf{r}_{i,\,k}\cdot\textbf{r}_{i,\,k+m}\\
       &=S_{ii,\,m}-2\mathrm{R}_{ii,\,m}

    where :math:`i` is the species index, :math:`m` is the index
    corresponding to time lag :math:`\tau`, :math:`\mathrm{R}_{ii,\,m}`
    is the autocorrelation of :math:`\mathbf{r}_i(t)`, and :math:`S_m`
    is evaluated using the recursive relation

    .. math::

       \begin{gather*}
         D_{ii,\,m}=\textbf{r}_{i,\,m}^2\\
         Q_{ii,\,-1}=2\sum_{k=0}^{N_t-1}D_{ii,\,k}\\
         Q_{ii,\,m}=Q_{ii,\,m-1}-D_{ii,\,m-1}-D_{ii,\,N_t-m}\\
         S_{ii,\,m}=\frac{Q_{ii,\,m}}{N_t-m}
       \end{gather*}

    Similarly, the CD for two sets of positions :math:`\mathbf{r}_i(t)`
    and :math:`\mathbf{r}_j(t)` is computed using

    .. math::

       \mathrm{CD}_{ij,m}=S_{ij,\,m}-2\mathrm{R}_{ij,\,m}

    where :math:`\mathrm{R}_{ij,\,m}` is the cross-correlation of
    :math:`\mathbf{r}_i(t)` and :math:`\mathbf{r}_j(t)`, and
    :math:`S_{ij,\,m}` is evaluated using the recursive relation

    .. math::

       \begin{gather*}
         D_{ij,\,m}=\textbf{r}_{i,\,m}\cdot\textbf{r}_{j,\,m}\\
         Q_{ij,\,-1}=2\sum_{k=0}^{N_t-1}D_{ij,\,k}\\
         Q_{ij,\,m}=Q_{ij,\,m-1}-D_{ij,\,m-1}-D_{ij,\,N_t-m}\\
         S_{ij,\,m}=\frac{Q_{ij,\,m}}{N_t-m}
       \end{gather*}

    .. note::

       To evaluate the sum in the expression used to calculate the
       Onsager transport coefficients [3]_

       .. math::

          L_{ij}=\frac{1}{6k_\mathrm{B}T}\lim_{t\rightarrow\infty}
          \frac{d}{d\tau}\left\langle
          \sum_{\alpha=1}^{N_i}[\mathbf{r}_{i,\,\alpha}(t_0+\tau)
          -\mathbf{r}_{i,\,\alpha}(t_0)]\cdot
          \sum_{\beta=1}^{N_j}[\mathbf{r}_{j,\,\beta}(t_0+\tau)
          -\mathbf{r}_{j,\,\beta(t_0)}]\right\rangle

       `r_i` and `r_j` should be summed over all entities before being
       passed to this function.

    Parameters
    ----------
    r_i : `numpy.ndarray`, positional-only
        Time evolution of individual or averaged positions for :math:`N`
        entities over :math:`N_\mathrm{b}` blocks of :math:`N_t` times
        each.

        **Shape**: :math:`(N_t,\,3)`, :math:`(N_t,\,N,\,3)`,
        :math:`(N_\mathrm{b},\,N_t,\,3)`, or
        :math:`(N_\mathrm{b},\,N_t,\,N,\,3)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    r_j : `numpy.ndarray`, positional-only, optional
        Time evolution of individual or averaged positions for another
        :math:`N` entities over :math:`N_\mathrm{b}` blocks of
        :math:`N_t` times each.

        **Shape**: Same as `r_i`.

        **Reference unit**: :math:`\mathrm{Å}`.

    axis : `int`, optional
        Axis along which time evolves. If not specified, the axis is
        determined automatically using the shape of `r_i`.

    average : `bool`, keyword-only, default: :code:`True`
        Determines whether the MSD/CD is averaged over all entities.
        Only available if `r_i` and `r_j` contain information for
        multiple entities.

    Returns
    -------
    disp : `numpy.ndarray`
        MSD or CD.

        **Shape**: Same as the shape of `r_i`, except with the last axis
        removed. If :code:`average=True`, the axis indexing the 
        :math:`N` entities is also removed.

        **Reference unit**: :math:`\text{Å}^2`.

    References
    ----------
    .. [1] Kneller, G. R.; Keiner, V.; Kneller, M.; Schiller, M.
       NMOLDYN: A Program Package for a Neutron Scattering Oriented
       Analysis of Molecular Dynamics Simulations. *Computer Physics
       Communications* **1995**, *91* (1–3), 191–214.
       https://doi.org/10.1016/0010-4655(95)00048-K.

    .. [2] Calandrini, V.; Pellegrini, E.; Calligari, P.; Hinsen, K.;
       Kneller, G. R. NMoldyn - Interfacing Spectroscopic Experiments,
       Molecular Dynamics Simulations and Models for Time Correlation
       Functions. *JDN* **2011**, *12*, 201–232.
       https://doi.org/10.1051/sfn/201112010.

    .. [3] Fong, K. D.; Self, J.; McCloskey, B. D.; Persson, K. A.
       Onsager Transport Coefficients and Transference Numbers in
       Polyelectrolyte Solutions and Polymerized Ionic Liquids.
       *Macromolecules* **2020**, *53* (21), 9503–9512.
       https://doi.org/10.1021/acs.macromol.0c02001.
    """

    # Ensure arrays have valid shapes
    r_i = np.asarray(r_i)
    if r_i.size == 0:
        raise ValueError("The position arrays cannot be empty.")
    if r_i.shape[-1] != 3:
        emsg = "The position arrays must have three components in " "the last axis."
        raise ValueError(emsg)
    ndim = r_i.ndim
    if not 2 <= ndim <= 4:
        emsg = "The position arrays must be two-, three-, or four-" "dimensional."
        raise ValueError(emsg)
    if r_j is not None:
        r_j = np.asarray(r_j)
        if r_i.shape != r_j.shape:
            raise ValueError("The position arrays must have the same shape.")

    # Check or set axis along which to compute the MSD/CD
    if axis is None:
        if ndim == 4:
            axis = 1
        else:
            axis = 0
            if ndim == 3:
                emsg = (
                    "The axis along which to compute the MSD/CD "
                    "was not specified and is ambiguous for a "
                    "three-dimensional array. As such, it has been "
                    "set to the first axis by default."
                )
                warnings.warn(emsg)
    elif axis not in {0, 1}:
        emsg = "The MSD/CD can only be computed along the first or " "second axis."
        raise ValueError(emsg)

    # Get intermediate quantities required for the MSD/CD calculation
    R_ij = correlation_fft(r_i, r_j, axis, average=False, double=True, vector=True)
    D_ij = (r_i * (r_i if r_j is None else r_j)).sum(axis=-1)

    N_t = r_i.shape[axis]
    if ndim - axis == 3:

        # Calculate the MSD/CD for each entity
        if not average:
            stack = np.vstack if ndim == 3 else np.hstack
            shape = np.asarray(r_i.shape[:-1])
            mask = np.ones_like(shape, dtype=bool)
            mask[axis] = False
            D_k = stack((D_ij, np.expand_dims(np.zeros(shape[mask]), axis)))
            if axis:
                Q_ij = 2 * D_k.sum(axis=axis, keepdims=True) * np.ones(
                    (1, N_t, 1)
                ) - np.cumsum(
                    D_k[:, np.arange(-1, N_t - 1)] + D_k[:, N_t:0:-1], axis=axis
                )
            else:
                Q_ij = 2 * D_k.sum(axis=axis) * np.ones((N_t, 1)) - np.cumsum(
                    D_k[np.arange(-1, N_t - 1)] + D_k[N_t:0:-1], axis=axis
                )
            return Q_ij / np.arange(N_t, 0, -1)[:, None] - R_ij

        # Average the intermediate quantities over all particles
        R_ij = R_ij.mean(axis=ndim - 2)
        D_ij = D_ij.mean(axis=ndim - 2)

    # Calculate the averaged MSD/CD
    if axis:
        Q_ij = 2 * D_ij.sum(axis=axis, keepdims=True) * np.ones((1, N_t)) - np.insert(
            np.cumsum(D_ij[:, : N_t - 1] + D_ij[:, N_t - 1 : 0 : -1], axis=axis),
            0,
            0,
            axis=axis,
        )
    else:
        Q_ij = 2 * D_ij.sum() * np.ones(N_t) - np.insert(
            np.cumsum(D_ij[: N_t - 1] + D_ij[N_t - 1 : 0 : -1]), 0, 0
        )
    return Q_ij / np.arange(N_t, 0, -1) - R_ij


def msd_shift(
    r_i: np.ndarray[float],
    r_j: np.ndarray[float] = None,
    /,
    axis: int = None,
    *,
    average: bool = True,
) -> np.ndarray[float]:
    r"""
    Evaluates the mean squared displacements (MSD) or the analogous
    cross displacements (CD) of positions :math:`\mathbf{r}_i(t)` and
    :math:`\mathbf{r}_j(t)` using the Einstein relation.

    For a set of positions :math:`\mathbf{r}_i(t)`, the MSD is defined
    as

    .. math::

        \mathrm{MSD}_i(\tau)=\left\langle[\textbf{r}_i(t_0+\tau)
        -\textbf{r}_i(t_0)]^2\right\rangle
        =\dfrac{1}{N_\tau}\sum_{k=1}^{N_\tau}[\textbf{r}_i(t_k+\tau)
        -\textbf{r}_i(t_k)]^2

    where :math:`\tau` is the time lag, :math:`t_j` is an arbitrary
    reference time, and :math:`N_\tau` is the number of possible
    reference times.

    Similarly, the CD for two sets of positions :math:`\mathbf{r}_i(t)`
    and :math:`\mathbf{r}_j(t)` is defined as

    .. math::

       \mathrm{CD}_{ij}(\tau)&=\langle
       [\textbf{r}_i(t_0+\tau)-\textbf{r}_i(t_0)]\cdot
       [\textbf{r}_j(t_0+\tau)-\textbf{r}_j(t_0)]\rangle\\
       &=\dfrac{1}{N_\tau}\sum_{k=1}^{N_\tau}
       [\textbf{r}_i(t_k+\tau)-\textbf{r}_i(t_k))\cdot
       (\textbf{r}_j(t_k+\tau)-\textbf{r}_j(t_k)]

    To minimize statistical noise, the MSD/CD is calculated for and
    averaged over all possible reference times :math:`t_0`.

    .. note::

       To evaluate the sum in the expression used to calculate the
       Onsager transport coefficients [1]_

       .. math::

          L_{ij}=\frac{1}{6k_\mathrm{B}T}\lim_{t\rightarrow\infty}
          \frac{d}{d\tau}\left\langle
          \sum_{\alpha=1}^{N_i}[\mathbf{r}_{i,\,\alpha}(t_0+\tau)
          -\mathbf{r}_{i,\,\alpha}(t_0)]\cdot
          \sum_{\beta=1}^{N_j}[\mathbf{r}_{j,\,\beta}(t_0+\tau)
          -\mathbf{r}_{j,\,\beta(t_0)}]\right\rangle

       `r_i` and `r_j` should be summed over all entities before being
       passed to this function.

    Parameters
    ----------
    r_i : `numpy.ndarray`, positional-only
        Time evolution of individual or averaged positions for :math:`N`
        entities over :math:`N_\mathrm{b}` blocks of :math:`N_t` times
        each.

        **Shape**: :math:`(N_t,\,3)`, :math:`(N_t,\,N,\,3)`,
        :math:`(N_\mathrm{b},\,N_t,\,3)`, or
        :math:`(N_\mathrm{b},\,N_t,\,N,\,3)`.

        **Reference unit**: :math:`\mathrm{Å}`.

    r_j : `numpy.ndarray`, positional-only, optional
        Time evolution of individual or averaged positions for another
        :math:`N` entities over :math:`N_\mathrm{b}` blocks of
        :math:`N_t` times each.

        **Shape**: Same as `r_i`.

        **Reference unit**: :math:`\mathrm{Å}`.

    axis : `int`, optional
        Axis along which time evolves. If not specified, the axis is
        determined automatically using the shape of `r_i`.

    average : `bool`, keyword-only, default: :code:`True`
        Determines whether the MSD/CD is averaged over all entities.
        Only available if `r_i` and `r_j` contain information for
        multiple entities.

    Returns
    -------
    disp : `numpy.ndarray`
        MSD or CD.

        **Shape**: Same as the shape of `r_i`, except with the last axis
        removed. If :code:`average=True`, the axis indexing the
        :math:`N` entities is also removed.

        **Reference unit**: :math:`\text{Å}^2`.

    References
    ----------
    .. [1] Fong, K. D.; Self, J.; McCloskey, B. D.; Persson, K. A.
       Onsager Transport Coefficients and Transference Numbers in
       Polyelectrolyte Solutions and Polymerized Ionic Liquids.
       *Macromolecules* **2020**, *53* (21), 9503–9512.
       https://doi.org/10.1021/acs.macromol.0c02001.
    """

    # Ensure arrays have valid shapes
    r_i = np.asarray(r_i)
    if r_i.size == 0:
        raise ValueError("The position arrays cannot be empty.")
    if r_i.shape[-1] != 3:
        emsg = "The position arrays must have three components in " "the last axis."
        raise ValueError(emsg)
    ndim = r_i.ndim
    if not 2 <= ndim <= 4:
        emsg = "The position arrays must be two-, three-, or four-" "dimensional."
        raise ValueError(emsg)
    if r_j is not None:
        r_j = np.asarray(r_j)
        if r_i.shape != r_j.shape:
            raise ValueError("The position arrays must have the same shape.")

    # Check or set axis along which to compute the MSD/CD
    if axis is None:
        if ndim == 4:
            axis = 1
        else:
            axis = 0
            if ndim == 3:
                emsg = (
                    "The axis along which to compute the MSD/CD "
                    "was not specified and is ambiguous for a "
                    "three-dimensional array. As such, it has been "
                    "set to the first axis by default."
                )
                warnings.warn(emsg)
    elif axis not in {0, 1}:
        emsg = "The MSD/CD can only be computed along the first or " "second axis."
        raise ValueError(emsg)

    # Calculate the MSD/CD for each entity
    N_t = r_i.shape[axis]
    if r_j is None:
        if axis:
            disp = np.stack(
                [
                    ((r_i[:, : -i if i else None] - r_i[:, i:]) ** 2)
                    .sum(axis=-1)
                    .mean(axis=axis)
                    for i in range(N_t)
                ],
                axis=1,
            )
        else:
            disp = np.stack(
                [
                    ((r_i[: -i if i else None] - r_i[i:]) ** 2)
                    .sum(axis=-1)
                    .mean(axis=axis)
                    for i in range(N_t)
                ]
            )
    else:
        if axis:
            disp = np.stack(
                [
                    np.einsum(
                        "bt...d,bt...d->bt...",
                        r_i[:, : -i if i else None] - r_i[:, i:],
                        r_j[:, : -i if i else None] - r_j[:, i:],
                    ).mean(axis=axis)
                    for i in range(N_t)
                ],
                axis=1,
            )
        else:
            disp = np.stack(
                [
                    np.einsum(
                        "t...d,t...d->t...",
                        r_i[: -i if i else None] - r_i[i:],
                        r_j[: -i if i else None] - r_j[i:],
                    ).mean(axis=axis)
                    for i in range(N_t)
                ]
            )

    # Average over all entities, if desired
    if ndim - axis == 3 and average:
        disp = disp.mean(axis=ndim - 2)

    return disp
